import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class InverseCross3D(_TileBase):
    """Simple inverse crosstile to tile with linear-quadratic branches and
    a trilinear center spline.

    Class that provides necessary functions to create inverse microtile,
    that can be used to describe the domain surrounding cross-tile
    microstructure.

    .. raw:: html

        <p><a href="../_static/InverseCross3D.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/InverseCross3D.html" />

    """  # noqa: E501

    _dim = 3
    _para_dim = 3
    _evaluation_points = _np.array(
        [
            [0.0, 0.5, 0.5],
            [1.0, 0.5, 0.5],
            [0.5, 0.0, 0.5],
            [0.5, 1.0, 0.5],
            [0.5, 0.5, 0.0],
            [0.5, 0.5, 1.0],
        ]
    )
    _n_info_per_eval_point = 1

    def _closing_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        closure=None,
        boundary_width=0.1,
        filling_height=0.5,
        seperator_distance=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a closing tile to match with closed surface.

        Parameters
        ----------
        parameters : tuple(np.ndarray)
          radii of fitting cylinder at evaluation points
        boundary_width : float
          with of the boundary surrounding branch
        filling_height : float
          portion of the height that is filled in parametric domain
        seperator_distance : float
          Describes the position of the separator layer in the control point
          domain
        closure : str
          parametric dimension that needs to be closed.
          Must be one of {"z_min", "z_max"}

        Returns
        -------
        list_of_splines : list
        derivatives: list<list<splines>> / None
        """
        # Check parameters
        if closure is None:
            raise ValueError("No closing direction given")

        # Set default values
        if seperator_distance is None:
            seperator_distance = 0.3

        if parameters is None:
            self._logd("Tile request is not parametrized, setting default 0.2")
            parameters = (
                _np.ones(
                    (
                        self._evaluation_points.shape[0],
                        self._n_info_per_eval_point,
                    )
                )
                * 0.2
            )

        self.check_params(parameters)

        if parameter_sensitivities is not None:
            raise NotImplementedError(
                "Derivatives are not implemented for this tile yet"
            )

        if not (_np.all(parameters > 0) and _np.all(parameters < 0.5)):
            raise ValueError("Thickness out of range (0, .5)")

        if not (0.0 < float(boundary_width) < 0.5):
            raise ValueError("Boundary Width is out of range")

        if not (0.0 < float(filling_height) < 1.0):
            raise ValueError("Filling must  be in (0,1)")

        # Precompute auxiliary values
        inv_filling_height = 1.0 - filling_height
        ctps_mid_height_top = (1 + filling_height) * 0.5
        ctps_mid_height_bottom = 1.0 - ctps_mid_height_top
        center_width = 1.0 - 2 * boundary_width
        r_center = center_width * 0.5
        half_r_center = (r_center + 0.5) * 0.5
        aux_column_width = 0.5 - 2 * (0.5 - seperator_distance)

        spline_list = []
        if closure == "z_min":
            branch_thickness = parameters.flatten()[5]
            branch_neighbor_x_min_ctps = _np.array(
                [
                    [-0.5, -r_center, filling_height],
                    [-half_r_center, -r_center, filling_height],
                    [-r_center, -r_center, filling_height],
                    [-0.5, r_center, filling_height],
                    [-half_r_center, r_center, filling_height],
                    [-r_center, r_center, filling_height],
                    [-0.5, -aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [-0.5, aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [-0.5, -aux_column_width, 1.0],
                    [-seperator_distance, -aux_column_width, 1.0],
                    [-branch_thickness, -branch_thickness, 1.0],
                    [-0.5, aux_column_width, 1.0],
                    [-seperator_distance, aux_column_width, 1.0],
                    [-branch_thickness, branch_thickness, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_min_ctps,
                )
            )

            branch_neighbor_x_max_ctps = _np.array(
                [
                    [r_center, -r_center, filling_height],
                    [half_r_center, -r_center, filling_height],
                    [0.5, -r_center, filling_height],
                    [r_center, r_center, filling_height],
                    [half_r_center, r_center, filling_height],
                    [0.5, r_center, filling_height],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_top],
                    [branch_thickness, -branch_thickness, 1.0],
                    [seperator_distance, -aux_column_width, 1.0],
                    [0.5, -aux_column_width, 1.0],
                    [branch_thickness, branch_thickness, 1.0],
                    [seperator_distance, aux_column_width, 1.0],
                    [0.5, aux_column_width, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_max_ctps,
                )
            )

            branch_neighbor_y_min_ctps = _np.array(
                [
                    [-r_center, -0.5, filling_height],
                    [r_center, -0.5, filling_height],
                    [-r_center, -half_r_center, filling_height],
                    [r_center, -half_r_center, filling_height],
                    [-r_center, -r_center, filling_height],
                    [r_center, -r_center, filling_height],
                    [-aux_column_width, -0.5, ctps_mid_height_top],
                    [aux_column_width, -0.5, ctps_mid_height_top],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [-aux_column_width, -0.5, 1.0],
                    [aux_column_width, -0.5, 1.0],
                    [-aux_column_width, -seperator_distance, 1.0],
                    [aux_column_width, -seperator_distance, 1.0],
                    [-branch_thickness, -branch_thickness, 1.0],
                    [branch_thickness, -branch_thickness, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_min_ctps,
                )
            )

            branch_neighbor_y_max_ctps = _np.array(
                [
                    [-r_center, r_center, filling_height],
                    [r_center, r_center, filling_height],
                    [-r_center, half_r_center, filling_height],
                    [r_center, half_r_center, filling_height],
                    [-r_center, 0.5, filling_height],
                    [r_center, 0.5, filling_height],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [-aux_column_width, 0.5, ctps_mid_height_top],
                    [aux_column_width, 0.5, ctps_mid_height_top],
                    [-branch_thickness, branch_thickness, 1.0],
                    [branch_thickness, branch_thickness, 1.0],
                    [-aux_column_width, seperator_distance, 1.0],
                    [aux_column_width, seperator_distance, 1.0],
                    [-aux_column_width, 0.5, 1.0],
                    [aux_column_width, 0.5, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_max_ctps,
                )
            )

            branch_x_min_y_min_ctps = _np.array(
                [
                    [-0.5, -0.5, filling_height],
                    [-half_r_center, -0.5, filling_height],
                    [-r_center, -0.5, filling_height],
                    [-0.5, -half_r_center, filling_height],
                    [-half_r_center, -half_r_center, filling_height],
                    [-r_center, -half_r_center, filling_height],
                    [-0.5, -r_center, filling_height],
                    [-half_r_center, -r_center, filling_height],
                    [-r_center, -r_center, filling_height],
                    [-0.5, -0.5, ctps_mid_height_top],
                    [-seperator_distance, -0.5, ctps_mid_height_top],
                    [-aux_column_width, -0.5, ctps_mid_height_top],
                    [-0.5, -seperator_distance, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [-0.5, -aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_top,
                    ],
                    [-0.5, -0.5, 1.0],
                    [-seperator_distance, -0.5, 1.0],
                    [-aux_column_width, -0.5, 1.0],
                    [-0.5, -seperator_distance, 1.0],
                    [-seperator_distance, -seperator_distance, 1.0],
                    [-aux_column_width, -seperator_distance, 1.0],
                    [-0.5, -aux_column_width, 1.0],
                    [-seperator_distance, -aux_column_width, 1.0],
                    [-branch_thickness, -branch_thickness, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_min_ctps
                )
            )

            branch_x_min_y_max_ctps = _np.array(
                [
                    [-0.5, r_center, filling_height],
                    [-half_r_center, r_center, filling_height],
                    [-r_center, r_center, filling_height],
                    [-0.5, half_r_center, filling_height],
                    [-half_r_center, half_r_center, filling_height],
                    [-r_center, half_r_center, filling_height],
                    [-0.5, 0.5, filling_height],
                    [-half_r_center, 0.5, filling_height],
                    [-r_center, 0.5, filling_height],
                    [-0.5, aux_column_width, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [-branch_thickness, branch_thickness, ctps_mid_height_top],
                    [-0.5, seperator_distance, ctps_mid_height_top],
                    [
                        -seperator_distance,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [-0.5, 0.5, ctps_mid_height_top],
                    [-seperator_distance, 0.5, ctps_mid_height_top],
                    [-aux_column_width, 0.5, ctps_mid_height_top],
                    [-0.5, aux_column_width, 1.0],
                    [-seperator_distance, aux_column_width, 1.0],
                    [-branch_thickness, branch_thickness, 1.0],
                    [-0.5, seperator_distance, 1.0],
                    [-seperator_distance, seperator_distance, 1.0],
                    [-aux_column_width, seperator_distance, 1.0],
                    [-0.5, 0.5, 1.0],
                    [-seperator_distance, 0.5, 1.0],
                    [-aux_column_width, 0.5, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_max_ctps
                )
            )

            branch_x_max_y_min_ctps = _np.array(
                [
                    [r_center, -0.5, filling_height],
                    [half_r_center, -0.5, filling_height],
                    [0.5, -0.5, filling_height],
                    [r_center, -half_r_center, filling_height],
                    [half_r_center, -half_r_center, filling_height],
                    [0.5, -half_r_center, filling_height],
                    [r_center, -r_center, filling_height],
                    [half_r_center, -r_center, filling_height],
                    [0.5, -r_center, filling_height],
                    [aux_column_width, -0.5, ctps_mid_height_top],
                    [seperator_distance, -0.5, ctps_mid_height_top],
                    [0.5, -0.5, ctps_mid_height_top],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [0.5, -seperator_distance, ctps_mid_height_top],
                    [branch_thickness, -branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_top],
                    [aux_column_width, -0.5, 1.0],
                    [seperator_distance, -0.5, 1.0],
                    [0.5, -0.5, 1.0],
                    [aux_column_width, -seperator_distance, 1.0],
                    [seperator_distance, -seperator_distance, 1.0],
                    [0.5, -seperator_distance, 1.0],
                    [branch_thickness, -branch_thickness, 1.0],
                    [seperator_distance, -aux_column_width, 1.0],
                    [0.5, -aux_column_width, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_min_ctps
                )
            )

            branch_x_max_y_max_ctps = _np.array(
                [
                    [r_center, r_center, filling_height],
                    [half_r_center, r_center, filling_height],
                    [0.5, r_center, filling_height],
                    [r_center, half_r_center, filling_height],
                    [half_r_center, half_r_center, filling_height],
                    [0.5, half_r_center, filling_height],
                    [r_center, 0.5, filling_height],
                    [half_r_center, 0.5, filling_height],
                    [0.5, 0.5, filling_height],
                    [branch_thickness, branch_thickness, ctps_mid_height_top],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_top,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_top],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [
                        seperator_distance,
                        seperator_distance,
                        ctps_mid_height_top,
                    ],
                    [0.5, seperator_distance, ctps_mid_height_top],
                    [aux_column_width, 0.5, ctps_mid_height_top],
                    [seperator_distance, 0.5, ctps_mid_height_top],
                    [0.5, 0.5, ctps_mid_height_top],
                    [branch_thickness, branch_thickness, 1.0],
                    [seperator_distance, aux_column_width, 1.0],
                    [0.5, aux_column_width, 1.0],
                    [aux_column_width, seperator_distance, 1.0],
                    [seperator_distance, seperator_distance, 1.0],
                    [0.5, seperator_distance, 1.0],
                    [aux_column_width, 0.5, 1.0],
                    [seperator_distance, 0.5, 1.0],
                    [0.5, 0.5, 1.0],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_max_ctps
                )
            )

            return (spline_list, None)

        elif closure == "z_max":
            branch_thickness = parameters.flatten()[4]
            branch_neighbor_x_min_ctps = _np.array(
                [
                    [-0.5, -aux_column_width, 0.0],
                    [-seperator_distance, -aux_column_width, 0.0],
                    [-branch_thickness, -branch_thickness, 0.0],
                    [-0.5, aux_column_width, 0.0],
                    [-seperator_distance, aux_column_width, 0.0],
                    [-branch_thickness, branch_thickness, 0.0],
                    [-0.5, -aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, -r_center, inv_filling_height],
                    [-half_r_center, -r_center, inv_filling_height],
                    [-r_center, -r_center, inv_filling_height],
                    [-0.5, r_center, inv_filling_height],
                    [-half_r_center, r_center, inv_filling_height],
                    [-r_center, r_center, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_min_ctps,
                )
            )

            branch_neighbor_x_max_ctps = _np.array(
                [
                    [branch_thickness, -branch_thickness, 0.0],
                    [seperator_distance, -aux_column_width, 0.0],
                    [0.5, -aux_column_width, 0.0],
                    [branch_thickness, branch_thickness, 0.0],
                    [seperator_distance, aux_column_width, 0.0],
                    [0.5, aux_column_width, 0.0],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_bottom],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_bottom],
                    [r_center, -r_center, inv_filling_height],
                    [half_r_center, -r_center, inv_filling_height],
                    [0.5, -r_center, inv_filling_height],
                    [r_center, r_center, inv_filling_height],
                    [half_r_center, r_center, inv_filling_height],
                    [0.5, r_center, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 1, 2],
                    control_points=branch_neighbor_x_max_ctps,
                )
            )

            branch_neighbor_y_min_ctps = _np.array(
                [
                    [-aux_column_width, -0.5, 0.0],
                    [aux_column_width, -0.5, 0.0],
                    [-aux_column_width, -seperator_distance, 0.0],
                    [aux_column_width, -seperator_distance, 0.0],
                    [-branch_thickness, -branch_thickness, 0.0],
                    [branch_thickness, -branch_thickness, 0.0],
                    [-aux_column_width, -0.5, ctps_mid_height_bottom],
                    [aux_column_width, -0.5, ctps_mid_height_bottom],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-r_center, -0.5, inv_filling_height],
                    [r_center, -0.5, inv_filling_height],
                    [-r_center, -half_r_center, inv_filling_height],
                    [r_center, -half_r_center, inv_filling_height],
                    [-r_center, -r_center, inv_filling_height],
                    [r_center, -r_center, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_min_ctps,
                )
            )

            branch_neighbor_y_max_ctps = _np.array(
                [
                    [-branch_thickness, branch_thickness, 0.0],
                    [branch_thickness, branch_thickness, 0.0],
                    [-aux_column_width, seperator_distance, 0.0],
                    [aux_column_width, seperator_distance, 0.0],
                    [-aux_column_width, 0.5, 0.0],
                    [aux_column_width, 0.5, 0.0],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [-aux_column_width, 0.5, ctps_mid_height_bottom],
                    [aux_column_width, 0.5, ctps_mid_height_bottom],
                    [-r_center, r_center, inv_filling_height],
                    [r_center, r_center, inv_filling_height],
                    [-r_center, half_r_center, inv_filling_height],
                    [r_center, half_r_center, inv_filling_height],
                    [-r_center, 0.5, inv_filling_height],
                    [r_center, 0.5, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[1, 2, 2],
                    control_points=branch_neighbor_y_max_ctps,
                )
            )

            branch_x_min_y_min_ctps = _np.array(
                [
                    [-0.5, -0.5, 0.0],
                    [-seperator_distance, -0.5, 0.0],
                    [-aux_column_width, -0.5, 0.0],
                    [-0.5, -seperator_distance, 0.0],
                    [-seperator_distance, -seperator_distance, 0.0],
                    [-aux_column_width, -seperator_distance, 0.0],
                    [-0.5, -aux_column_width, 0.0],
                    [-seperator_distance, -aux_column_width, 0.0],
                    [-branch_thickness, -branch_thickness, 0.0],
                    [-0.5, -0.5, ctps_mid_height_bottom],
                    [-seperator_distance, -0.5, ctps_mid_height_bottom],
                    [-aux_column_width, -0.5, ctps_mid_height_bottom],
                    [-0.5, -seperator_distance, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, -aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, -0.5, inv_filling_height],
                    [-half_r_center, -0.5, inv_filling_height],
                    [-r_center, -0.5, inv_filling_height],
                    [-0.5, -half_r_center, inv_filling_height],
                    [-half_r_center, -half_r_center, inv_filling_height],
                    [-r_center, -half_r_center, inv_filling_height],
                    [-0.5, -r_center, inv_filling_height],
                    [-half_r_center, -r_center, inv_filling_height],
                    [-r_center, -r_center, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_min_ctps
                )
            )

            branch_x_max_y_max_ctps = _np.array(
                [
                    [branch_thickness, branch_thickness, 0.0],
                    [seperator_distance, aux_column_width, 0.0],
                    [0.5, aux_column_width, 0.0],
                    [aux_column_width, seperator_distance, 0.0],
                    [seperator_distance, seperator_distance, 0.0],
                    [0.5, seperator_distance, 0.0],
                    [aux_column_width, 0.5, 0.0],
                    [seperator_distance, 0.5, 0.0],
                    [0.5, 0.5, 0.0],
                    [
                        branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, aux_column_width, ctps_mid_height_bottom],
                    [
                        aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, seperator_distance, ctps_mid_height_bottom],
                    [aux_column_width, 0.5, ctps_mid_height_bottom],
                    [seperator_distance, 0.5, ctps_mid_height_bottom],
                    [0.5, 0.5, ctps_mid_height_bottom],
                    [r_center, r_center, inv_filling_height],
                    [half_r_center, r_center, inv_filling_height],
                    [0.5, r_center, inv_filling_height],
                    [r_center, half_r_center, inv_filling_height],
                    [half_r_center, half_r_center, inv_filling_height],
                    [0.5, half_r_center, inv_filling_height],
                    [r_center, 0.5, inv_filling_height],
                    [half_r_center, 0.5, inv_filling_height],
                    [0.5, 0.5, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_max_ctps
                )
            )

            branch_x_max_y_min_ctps = _np.array(
                [
                    [aux_column_width, -0.5, 0.0],
                    [seperator_distance, -0.5, 0.0],
                    [0.5, -0.5, 0.0],
                    [aux_column_width, -seperator_distance, 0.0],
                    [seperator_distance, -seperator_distance, 0.0],
                    [0.5, -seperator_distance, 0.0],
                    [branch_thickness, -branch_thickness, 0.0],
                    [seperator_distance, -aux_column_width, 0.0],
                    [0.5, -aux_column_width, 0.0],
                    [aux_column_width, -0.5, ctps_mid_height_bottom],
                    [seperator_distance, -0.5, ctps_mid_height_bottom],
                    [0.5, -0.5, ctps_mid_height_bottom],
                    [
                        aux_column_width,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        -seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, -seperator_distance, ctps_mid_height_bottom],
                    [
                        branch_thickness,
                        -branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [
                        seperator_distance,
                        -aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [0.5, -aux_column_width, ctps_mid_height_bottom],
                    [r_center, -0.5, inv_filling_height],
                    [half_r_center, -0.5, inv_filling_height],
                    [0.5, -0.5, inv_filling_height],
                    [r_center, -half_r_center, inv_filling_height],
                    [half_r_center, -half_r_center, inv_filling_height],
                    [0.5, -half_r_center, inv_filling_height],
                    [r_center, -r_center, inv_filling_height],
                    [half_r_center, -r_center, inv_filling_height],
                    [0.5, -r_center, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_max_y_min_ctps
                )
            )

            branch_x_min_y_max_ctps = _np.array(
                [
                    [-0.5, aux_column_width, 0.0],
                    [-seperator_distance, aux_column_width, 0.0],
                    [-branch_thickness, branch_thickness, 0.0],
                    [-0.5, seperator_distance, 0.0],
                    [-seperator_distance, seperator_distance, 0.0],
                    [-aux_column_width, seperator_distance, 0.0],
                    [-0.5, 0.5, 0.0],
                    [-seperator_distance, 0.5, 0.0],
                    [-aux_column_width, 0.5, 0.0],
                    [-0.5, aux_column_width, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        aux_column_width,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -branch_thickness,
                        branch_thickness,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, seperator_distance, ctps_mid_height_bottom],
                    [
                        -seperator_distance,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [
                        -aux_column_width,
                        seperator_distance,
                        ctps_mid_height_bottom,
                    ],
                    [-0.5, 0.5, ctps_mid_height_bottom],
                    [-seperator_distance, 0.5, ctps_mid_height_bottom],
                    [-aux_column_width, 0.5, ctps_mid_height_bottom],
                    [-0.5, r_center, inv_filling_height],
                    [-half_r_center, r_center, inv_filling_height],
                    [-r_center, r_center, inv_filling_height],
                    [-0.5, half_r_center, inv_filling_height],
                    [-half_r_center, half_r_center, inv_filling_height],
                    [-r_center, half_r_center, inv_filling_height],
                    [-0.5, 0.5, inv_filling_height],
                    [-half_r_center, 0.5, inv_filling_height],
                    [-r_center, 0.5, inv_filling_height],
                ]
            ) + _np.array([0.5, 0.5, 0.0])

            spline_list.append(
                _Bezier(
                    degrees=[2, 2, 2], control_points=branch_x_min_y_max_ctps
                )
            )

            return (spline_list, None)
        else:
            raise ValueError("Corner Type not supported")

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        seperator_distance=None,
        center_expansion=1.0,
        closure=None,
        **kwargs,  # noqa ARG002
    ):
        """Create an inverse microtile based on the parameters that describe
        the branch thicknesses.

        Thickness parameters are used to describe the inner radius of the
        outward facing branches

        Parameters
        ----------
        parameters : np.array
          only first entry is used, defines the internal radii of the
          branches
        seperator_distance : float
          Control point distance to separation layer of biquadratic degrees,
          determines the minimum branch thickness (defaults to 0.4)
        center_expansion : float
          thickness of center is expanded by a factor (default to 1.0), which
          determines the maximum branch thickness
        closure : str
          parametric dimension that needs to be closed.
          Must be one of {"z_min", "z_max"}

        Returns
        -------
        microtile_list : list(splines)
        """

        if not isinstance(center_expansion, float):
            raise ValueError("Invalid Type")

        if not ((center_expansion > 0.5) and (center_expansion < 1.5)):
            raise ValueError("Center Expansion must be in (.5, 1.5)")

        # Set default values
        if seperator_distance is None:
            seperator_distance = 0.3

        if center_expansion is None:
            center_expansion = 1.0

        # Check if all radii are in allowed range
        max_radius = min(0.5, (0.5 / center_expansion))
        max_radius = min(max_radius, seperator_distance)
        min_radius = max(0.5 - seperator_distance, 0)

        # set to default if nothing is given
        if parameters is None:
            self._logd("Setting branch thickness to default 0.2")
            parameters = (
                _np.ones(
                    (len(self._evaluation_points), self._n_info_per_eval_point)
                )
                * 0.2
            )

        if parameter_sensitivities is not None:
            raise NotImplementedError(
                "Derivatives are not implemented for this tile yet"
            )

        self.check_params(parameters)

        if _np.any(parameters < min_radius) or _np.any(
            parameters > max_radius
        ):
            raise ValueError(
                f"Radii must be in (0,{max_radius}) for "
                f"center_expansion {center_expansion}"
            )

        if closure is not None:
            return self._closing_tile(
                parameters=parameters,
                parameter_sensitivities=parameter_sensitivities,
                seperator_distance=seperator_distance,
                closure=closure,
                **kwargs,
            )

        [
            x_min_r,
            x_max_r,
            y_min_r,
            y_max_r,
            z_min_r,
            z_max_r,
        ] = parameters.flatten()

        # center radius
        center_r = _np.sum(parameters) / 6.0 * center_expansion

        # Auxiliary values for smooothing (mid-branch thickness)
        [
            aux_x_min,
            aux_x_max,
            aux_y_min,
            aux_y_max,
            aux_z_min,
            aux_z_max,
        ] = _np.minimum(parameters.ravel(), center_r)
        # Branch midlength
        hd_center = 0.5 * (0.5 + center_r)
        aux_column_width = 0.5 - 2 * (0.5 - seperator_distance)

        # Init return type
        spline_list = []

        # Start with branch interconnections
        x_min_y_min = _np.array(
            [
                [-0.5, -0.5, -aux_column_width],
                [-seperator_distance, -0.5, -aux_column_width],
                [-y_min_r, -0.5, -y_min_r],
                [-0.5, -seperator_distance, -aux_column_width],
                [-hd_center, -hd_center, -aux_column_width],
                [-aux_y_min, -hd_center, -aux_y_min],
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                [-center_r, -center_r, -center_r],
                [-0.5, -0.5, aux_column_width],
                [-seperator_distance, -0.5, aux_column_width],
                [-y_min_r, -0.5, y_min_r],
                [-0.5, -seperator_distance, aux_column_width],
                [-hd_center, -hd_center, aux_column_width],
                [-aux_y_min, -hd_center, aux_y_min],
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                [-center_r, -center_r, center_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 1], control_points=x_min_y_min)
        )

        x_max_y_min = _np.array(
            [
                [y_min_r, -0.5, -y_min_r],
                [seperator_distance, -0.5, -aux_column_width],
                [0.5, -0.5, -aux_column_width],
                [aux_y_min, -hd_center, -aux_y_min],
                [hd_center, -hd_center, -aux_column_width],
                [0.5, -seperator_distance, -aux_column_width],
                [center_r, -center_r, -center_r],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
                [y_min_r, -0.5, y_min_r],
                [seperator_distance, -0.5, aux_column_width],
                [0.5, -0.5, aux_column_width],
                [aux_y_min, -hd_center, aux_y_min],
                [hd_center, -hd_center, aux_column_width],
                [0.5, -seperator_distance, aux_column_width],
                [center_r, -center_r, center_r],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 1], control_points=x_max_y_min)
        )

        x_min_y_max = _np.array(
            [
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                [-center_r, center_r, -center_r],
                [-0.5, seperator_distance, -aux_column_width],
                [-hd_center, hd_center, -aux_column_width],
                [-aux_y_max, hd_center, -aux_y_max],
                [-0.5, 0.5, -aux_column_width],
                [-seperator_distance, 0.5, -aux_column_width],
                [-y_max_r, 0.5, -y_max_r],
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                [-center_r, center_r, center_r],
                [-0.5, seperator_distance, aux_column_width],
                [-hd_center, hd_center, aux_column_width],
                [-aux_y_max, hd_center, aux_y_max],
                [-0.5, 0.5, aux_column_width],
                [-seperator_distance, 0.5, aux_column_width],
                [-y_max_r, 0.5, y_max_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 1], control_points=x_min_y_max)
        )

        x_max_y_max = _np.array(
            [
                [center_r, center_r, -center_r],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
                [aux_y_max, hd_center, -aux_y_max],
                [hd_center, hd_center, -aux_column_width],
                [0.5, seperator_distance, -aux_column_width],
                [y_max_r, 0.5, -y_max_r],
                [seperator_distance, 0.5, -aux_column_width],
                [0.5, 0.5, -aux_column_width],
                [center_r, center_r, center_r],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
                [aux_y_max, hd_center, aux_y_max],
                [hd_center, hd_center, aux_column_width],
                [0.5, seperator_distance, aux_column_width],
                [y_max_r, 0.5, y_max_r],
                [seperator_distance, 0.5, aux_column_width],
                [0.5, 0.5, aux_column_width],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 1], control_points=x_max_y_max)
        )

        x_min_z_min = _np.array(
            [
                [-0.5, -aux_column_width, -0.5],
                [-seperator_distance, -aux_column_width, -0.5],
                [-z_min_r, -z_min_r, -0.5],
                [-0.5, aux_column_width, -0.5],
                [-seperator_distance, aux_column_width, -0.5],
                [-z_min_r, z_min_r, -0.5],
                [-0.5, -aux_column_width, -seperator_distance],
                [-hd_center, -aux_column_width, -hd_center],
                [-aux_z_min, -aux_z_min, -hd_center],
                [-0.5, aux_column_width, -seperator_distance],
                [-hd_center, aux_column_width, -hd_center],
                [-aux_z_min, aux_z_min, -hd_center],
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                [-center_r, -center_r, -center_r],
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                [-center_r, center_r, -center_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 1, 2], control_points=x_min_z_min)
        )

        x_max_z_min = _np.array(
            [
                [z_min_r, -z_min_r, -0.5],
                [seperator_distance, -aux_column_width, -0.5],
                [0.5, -aux_column_width, -0.5],
                [z_min_r, z_min_r, -0.5],
                [seperator_distance, aux_column_width, -0.5],
                [0.5, aux_column_width, -0.5],
                [aux_z_min, -aux_z_min, -hd_center],
                [hd_center, -aux_column_width, -hd_center],
                [0.5, -aux_column_width, -seperator_distance],
                [aux_z_min, aux_z_min, -hd_center],
                [hd_center, aux_column_width, -hd_center],
                [0.5, aux_column_width, -seperator_distance],
                [center_r, -center_r, -center_r],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
                [center_r, center_r, -center_r],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 1, 2], control_points=x_max_z_min)
        )

        x_min_z_max = _np.array(
            [
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                [-center_r, -center_r, center_r],
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                [-center_r, center_r, center_r],
                [-0.5, -aux_column_width, seperator_distance],
                [-hd_center, -aux_column_width, hd_center],
                [-aux_z_max, -aux_z_max, hd_center],
                [-0.5, aux_column_width, seperator_distance],
                [-hd_center, aux_column_width, hd_center],
                [-aux_z_max, aux_z_max, hd_center],
                [-0.5, -aux_column_width, 0.5],
                [-seperator_distance, -aux_column_width, 0.5],
                [-z_max_r, -z_max_r, 0.5],
                [-0.5, aux_column_width, 0.5],
                [-seperator_distance, aux_column_width, 0.5],
                [-z_max_r, z_max_r, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 1, 2], control_points=x_min_z_max)
        )

        x_max_z_max = _np.array(
            [
                [center_r, -center_r, center_r],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
                [center_r, center_r, center_r],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
                [aux_z_max, -aux_z_max, hd_center],
                [hd_center, -aux_column_width, hd_center],
                [0.5, -aux_column_width, seperator_distance],
                [aux_z_max, aux_z_max, hd_center],
                [hd_center, aux_column_width, hd_center],
                [0.5, aux_column_width, seperator_distance],
                [z_max_r, -z_max_r, 0.5],
                [seperator_distance, -aux_column_width, 0.5],
                [0.5, -aux_column_width, 0.5],
                [z_max_r, z_max_r, 0.5],
                [seperator_distance, aux_column_width, 0.5],
                [0.5, aux_column_width, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 1, 2], control_points=x_max_z_max)
        )

        y_min_z_min = _np.array(
            [
                [-aux_column_width, -0.5, -0.5],
                [aux_column_width, -0.5, -0.5],
                [-aux_column_width, -seperator_distance, -0.5],
                [aux_column_width, -seperator_distance, -0.5],
                [-z_min_r, -z_min_r, -0.5],
                [z_min_r, -z_min_r, -0.5],
                [-aux_column_width, -0.5, -seperator_distance],
                [aux_column_width, -0.5, -seperator_distance],
                [-aux_column_width, -hd_center, -hd_center],
                [aux_column_width, -hd_center, -hd_center],
                [-aux_z_min, -aux_z_min, -hd_center],
                [aux_z_min, -aux_z_min, -hd_center],
                [-y_min_r, -0.5, -y_min_r],
                [y_min_r, -0.5, -y_min_r],
                [-aux_y_min, -hd_center, -aux_y_min],
                [aux_y_min, -hd_center, -aux_y_min],
                [-center_r, -center_r, -center_r],
                [center_r, -center_r, -center_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[1, 2, 2], control_points=y_min_z_min)
        )

        y_max_z_min = _np.array(
            [
                [-z_min_r, z_min_r, -0.5],
                [z_min_r, z_min_r, -0.5],
                [-aux_column_width, seperator_distance, -0.5],
                [aux_column_width, seperator_distance, -0.5],
                [-aux_column_width, 0.5, -0.5],
                [aux_column_width, 0.5, -0.5],
                [-aux_z_min, aux_z_min, -hd_center],
                [aux_z_min, aux_z_min, -hd_center],
                [-aux_column_width, hd_center, -hd_center],
                [aux_column_width, hd_center, -hd_center],
                [-aux_column_width, 0.5, -seperator_distance],
                [aux_column_width, 0.5, -seperator_distance],
                [-center_r, center_r, -center_r],
                [center_r, center_r, -center_r],
                [-aux_y_max, hd_center, -aux_y_max],
                [aux_y_max, hd_center, -aux_y_max],
                [-y_max_r, 0.5, -y_max_r],
                [y_max_r, 0.5, -y_max_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[1, 2, 2], control_points=y_max_z_min)
        )

        y_min_z_max = _np.array(
            [
                [-y_min_r, -0.5, y_min_r],
                [y_min_r, -0.5, y_min_r],
                [-aux_y_min, -hd_center, aux_y_min],
                [aux_y_min, -hd_center, aux_y_min],
                [-center_r, -center_r, center_r],
                [center_r, -center_r, center_r],
                [-aux_column_width, -0.5, seperator_distance],
                [aux_column_width, -0.5, seperator_distance],
                [-aux_column_width, -hd_center, hd_center],
                [aux_column_width, -hd_center, hd_center],
                [-aux_z_max, -aux_z_max, hd_center],
                [aux_z_max, -aux_z_max, hd_center],
                [-aux_column_width, -0.5, 0.5],
                [aux_column_width, -0.5, 0.5],
                [-aux_column_width, -seperator_distance, 0.5],
                [aux_column_width, -seperator_distance, 0.5],
                [-z_max_r, -z_max_r, 0.5],
                [z_max_r, -z_max_r, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[1, 2, 2], control_points=y_min_z_max)
        )

        y_max_z_max = _np.array(
            [
                [-center_r, center_r, center_r],
                [center_r, center_r, center_r],
                [-aux_y_max, hd_center, aux_y_max],
                [aux_y_max, hd_center, aux_y_max],
                [-y_max_r, 0.5, y_max_r],
                [y_max_r, 0.5, y_max_r],
                [-aux_z_max, aux_z_max, hd_center],
                [aux_z_max, aux_z_max, hd_center],
                [-aux_column_width, hd_center, hd_center],
                [aux_column_width, hd_center, hd_center],
                [-aux_column_width, 0.5, seperator_distance],
                [aux_column_width, 0.5, seperator_distance],
                [-z_max_r, z_max_r, 0.5],
                [z_max_r, z_max_r, 0.5],
                [-aux_column_width, seperator_distance, 0.5],
                [aux_column_width, seperator_distance, 0.5],
                [-aux_column_width, 0.5, 0.5],
                [aux_column_width, 0.5, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[1, 2, 2], control_points=y_max_z_max)
        )

        x_min_y_min_z_min = _np.array(
            [
                [-0.5, -0.5, -0.5],
                [-seperator_distance, -0.5, -0.5],
                [-aux_column_width, -0.5, -0.5],
                [-0.5, -seperator_distance, -0.5],
                [-seperator_distance, -seperator_distance, -0.5],
                [-aux_column_width, -seperator_distance, -0.5],
                [-0.5, -aux_column_width, -0.5],
                [-seperator_distance, -aux_column_width, -0.5],
                [-z_min_r, -z_min_r, -0.5],
                [-0.5, -0.5, -seperator_distance],
                [-seperator_distance, -0.5, -seperator_distance],
                [-aux_column_width, -0.5, -seperator_distance],
                [-0.5, -seperator_distance, -seperator_distance],
                [-hd_center, -hd_center, -hd_center],
                [-aux_column_width, -hd_center, -hd_center],
                [-0.5, -aux_column_width, -seperator_distance],
                [-hd_center, -aux_column_width, -hd_center],
                [-aux_z_min, -aux_z_min, -hd_center],
                [-0.5, -0.5, -aux_column_width],
                [-seperator_distance, -0.5, -aux_column_width],
                [-y_min_r, -0.5, -y_min_r],
                [-0.5, -seperator_distance, -aux_column_width],
                [-hd_center, -hd_center, -aux_column_width],
                [-aux_y_min, -hd_center, -aux_y_min],
                [-0.5, -x_min_r, -x_min_r],
                [-hd_center, -aux_x_min, -aux_x_min],
                [-center_r, -center_r, -center_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_min_y_min_z_min)
        )

        x_max_y_min_z_min = _np.array(
            [
                [aux_column_width, -0.5, -0.5],
                [seperator_distance, -0.5, -0.5],
                [0.5, -0.5, -0.5],
                [aux_column_width, -seperator_distance, -0.5],
                [seperator_distance, -seperator_distance, -0.5],
                [0.5, -seperator_distance, -0.5],
                [z_min_r, -z_min_r, -0.5],
                [seperator_distance, -aux_column_width, -0.5],
                [0.5, -aux_column_width, -0.5],
                [aux_column_width, -0.5, -seperator_distance],
                [seperator_distance, -0.5, -seperator_distance],
                [0.5, -0.5, -seperator_distance],
                [aux_column_width, -hd_center, -hd_center],
                [hd_center, -hd_center, -hd_center],
                [0.5, -seperator_distance, -seperator_distance],
                [aux_z_min, -aux_z_min, -hd_center],
                [hd_center, -aux_column_width, -hd_center],
                [0.5, -aux_column_width, -seperator_distance],
                [y_min_r, -0.5, -y_min_r],
                [seperator_distance, -0.5, -aux_column_width],
                [0.5, -0.5, -aux_column_width],
                [aux_y_min, -hd_center, -aux_y_min],
                [hd_center, -hd_center, -aux_column_width],
                [0.5, -seperator_distance, -aux_column_width],
                [center_r, -center_r, -center_r],
                [hd_center, -aux_x_max, -aux_x_max],
                [0.5, -x_max_r, -x_max_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_max_y_min_z_min)
        )

        x_min_y_max_z_min = _np.array(
            [
                [-0.5, aux_column_width, -0.5],
                [-seperator_distance, aux_column_width, -0.5],
                [-z_min_r, z_min_r, -0.5],
                [-0.5, seperator_distance, -0.5],
                [-seperator_distance, seperator_distance, -0.5],
                [-aux_column_width, seperator_distance, -0.5],
                [-0.5, 0.5, -0.5],
                [-seperator_distance, 0.5, -0.5],
                [-aux_column_width, 0.5, -0.5],
                [-0.5, aux_column_width, -seperator_distance],
                [-hd_center, aux_column_width, -hd_center],
                [-aux_z_min, aux_z_min, -hd_center],
                [-0.5, seperator_distance, -seperator_distance],
                [-hd_center, hd_center, -hd_center],
                [-aux_column_width, hd_center, -hd_center],
                [-0.5, 0.5, -seperator_distance],
                [-seperator_distance, 0.5, -seperator_distance],
                [-aux_column_width, 0.5, -seperator_distance],
                [-0.5, x_min_r, -x_min_r],
                [-hd_center, aux_x_min, -aux_x_min],
                [-center_r, center_r, -center_r],
                [-0.5, seperator_distance, -aux_column_width],
                [-hd_center, hd_center, -aux_column_width],
                [-aux_y_max, hd_center, -aux_y_max],
                [-0.5, 0.5, -aux_column_width],
                [-seperator_distance, 0.5, -aux_column_width],
                [-y_max_r, 0.5, -y_max_r],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_min_y_max_z_min)
        )

        x_max_y_max_z_min = _np.array(
            [
                [z_min_r, z_min_r, -0.5],
                [seperator_distance, aux_column_width, -0.5],
                [0.5, aux_column_width, -0.5],
                [aux_column_width, seperator_distance, -0.5],
                [seperator_distance, seperator_distance, -0.5],
                [0.5, seperator_distance, -0.5],
                [aux_column_width, 0.5, -0.5],
                [seperator_distance, 0.5, -0.5],
                [0.5, 0.5, -0.5],
                [aux_z_min, aux_z_min, -hd_center],
                [hd_center, aux_column_width, -hd_center],
                [0.5, aux_column_width, -seperator_distance],
                [aux_column_width, hd_center, -hd_center],
                [hd_center, hd_center, -hd_center],
                [0.5, seperator_distance, -seperator_distance],
                [aux_column_width, 0.5, -seperator_distance],
                [seperator_distance, 0.5, -seperator_distance],
                [0.5, 0.5, -seperator_distance],
                [center_r, center_r, -center_r],
                [hd_center, aux_x_max, -aux_x_max],
                [0.5, x_max_r, -x_max_r],
                [aux_y_max, hd_center, -aux_y_max],
                [hd_center, hd_center, -aux_column_width],
                [0.5, seperator_distance, -aux_column_width],
                [y_max_r, 0.5, -y_max_r],
                [seperator_distance, 0.5, -aux_column_width],
                [0.5, 0.5, -aux_column_width],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_max_y_max_z_min)
        )

        x_min_y_min_z_max = _np.array(
            [
                [-0.5, -0.5, aux_column_width],
                [-seperator_distance, -0.5, aux_column_width],
                [-y_min_r, -0.5, y_min_r],
                [-0.5, -seperator_distance, aux_column_width],
                [-hd_center, -hd_center, aux_column_width],
                [-aux_y_min, -hd_center, aux_y_min],
                [-0.5, -x_min_r, x_min_r],
                [-hd_center, -aux_x_min, aux_x_min],
                [-center_r, -center_r, center_r],
                [-0.5, -0.5, seperator_distance],
                [-seperator_distance, -0.5, seperator_distance],
                [-aux_column_width, -0.5, seperator_distance],
                [-0.5, -seperator_distance, seperator_distance],
                [-hd_center, -hd_center, hd_center],
                [-aux_column_width, -hd_center, hd_center],
                [-0.5, -aux_column_width, seperator_distance],
                [-hd_center, -aux_column_width, hd_center],
                [-aux_z_max, -aux_z_max, hd_center],
                [-0.5, -0.5, 0.5],
                [-seperator_distance, -0.5, 0.5],
                [-aux_column_width, -0.5, 0.5],
                [-0.5, -seperator_distance, 0.5],
                [-seperator_distance, -seperator_distance, 0.5],
                [-aux_column_width, -seperator_distance, 0.5],
                [-0.5, -aux_column_width, 0.5],
                [-seperator_distance, -aux_column_width, 0.5],
                [-z_max_r, -z_max_r, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_min_y_min_z_max)
        )

        x_max_y_min_z_max = _np.array(
            [
                [y_min_r, -0.5, y_min_r],
                [seperator_distance, -0.5, aux_column_width],
                [0.5, -0.5, aux_column_width],
                [aux_y_min, -hd_center, aux_y_min],
                [hd_center, -hd_center, aux_column_width],
                [0.5, -seperator_distance, aux_column_width],
                [center_r, -center_r, center_r],
                [hd_center, -aux_x_max, aux_x_max],
                [0.5, -x_max_r, x_max_r],
                [aux_column_width, -0.5, seperator_distance],
                [seperator_distance, -0.5, seperator_distance],
                [0.5, -0.5, seperator_distance],
                [aux_column_width, -hd_center, hd_center],
                [hd_center, -hd_center, hd_center],
                [0.5, -seperator_distance, seperator_distance],
                [aux_z_max, -aux_z_max, hd_center],
                [hd_center, -aux_column_width, hd_center],
                [0.5, -aux_column_width, seperator_distance],
                [aux_column_width, -0.5, 0.5],
                [seperator_distance, -0.5, 0.5],
                [0.5, -0.5, 0.5],
                [aux_column_width, -seperator_distance, 0.5],
                [seperator_distance, -seperator_distance, 0.5],
                [0.5, -seperator_distance, 0.5],
                [z_max_r, -z_max_r, 0.5],
                [seperator_distance, -aux_column_width, 0.5],
                [0.5, -aux_column_width, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_max_y_min_z_max)
        )

        x_min_y_max_z_max = _np.array(
            [
                [-0.5, x_min_r, x_min_r],
                [-hd_center, aux_x_min, aux_x_min],
                [-center_r, center_r, center_r],
                [-0.5, seperator_distance, aux_column_width],
                [-hd_center, hd_center, aux_column_width],
                [-aux_y_max, hd_center, aux_y_max],
                [-0.5, 0.5, aux_column_width],
                [-seperator_distance, 0.5, aux_column_width],
                [-y_max_r, 0.5, y_max_r],
                [-0.5, aux_column_width, seperator_distance],
                [-hd_center, aux_column_width, hd_center],
                [-aux_z_max, aux_z_max, hd_center],
                [-0.5, seperator_distance, seperator_distance],
                [-hd_center, hd_center, hd_center],
                [-aux_column_width, hd_center, hd_center],
                [-0.5, 0.5, seperator_distance],
                [-seperator_distance, 0.5, seperator_distance],
                [-aux_column_width, 0.5, seperator_distance],
                [-0.5, aux_column_width, 0.5],
                [-seperator_distance, aux_column_width, 0.5],
                [-z_max_r, z_max_r, 0.5],
                [-0.5, seperator_distance, 0.5],
                [-seperator_distance, seperator_distance, 0.5],
                [-aux_column_width, seperator_distance, 0.5],
                [-0.5, 0.5, 0.5],
                [-seperator_distance, 0.5, 0.5],
                [-aux_column_width, 0.5, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_min_y_max_z_max)
        )

        x_max_y_max_z_max = _np.array(
            [
                [center_r, center_r, center_r],
                [hd_center, aux_x_max, aux_x_max],
                [0.5, x_max_r, x_max_r],
                [aux_y_max, hd_center, aux_y_max],
                [hd_center, hd_center, aux_column_width],
                [0.5, seperator_distance, aux_column_width],
                [y_max_r, 0.5, y_max_r],
                [seperator_distance, 0.5, aux_column_width],
                [0.5, 0.5, aux_column_width],
                [aux_z_max, aux_z_max, hd_center],
                [hd_center, aux_column_width, hd_center],
                [0.5, aux_column_width, seperator_distance],
                [aux_column_width, hd_center, hd_center],
                [hd_center, hd_center, hd_center],
                [0.5, seperator_distance, seperator_distance],
                [aux_column_width, 0.5, seperator_distance],
                [seperator_distance, 0.5, seperator_distance],
                [0.5, 0.5, seperator_distance],
                [z_max_r, z_max_r, 0.5],
                [seperator_distance, aux_column_width, 0.5],
                [0.5, aux_column_width, 0.5],
                [aux_column_width, seperator_distance, 0.5],
                [seperator_distance, seperator_distance, 0.5],
                [0.5, seperator_distance, 0.5],
                [aux_column_width, 0.5, 0.5],
                [seperator_distance, 0.5, 0.5],
                [0.5, 0.5, 0.5],
            ]
        ) + _np.array([0.5, 0.5, 0.5])

        spline_list.append(
            _Bezier(degrees=[2, 2, 2], control_points=x_max_y_max_z_max)
        )
        return (spline_list, None)



import os
import tempfile
from sys import version as python_version

import gustaf as gus
import numpy as np

import splinepy

try:
    import vedo  # noqa F401

    has_vedo = True
except ImportError:
    has_vedo = False


def test_svg_export(to_tmpf, are_stripped_lines_same):
    """
    Test export routine of svg module
    """
    if not has_vedo:
        return

    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 1], [1, 1], [0, 2], [1, 2]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bsp_el2.elevate_degrees(0)
    bsp_el2.elevate_degrees(1)
    nur_el3.elevate_degrees(0)
    nur_el3.elevate_degrees(1)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Set some show_options
    nur_el3.show_options["control_points"] = False
    bez_el0.show_options["c"] = "grey"
    bez_el0.show_options["alpha"] = 0.5
    bez_el0.show_options["control_mesh_lw"] = 0.2
    bez_el0.show_options["control_mesh_c"] = "orange"
    bez_el0.show_options["knot_c"] = "blue"
    bsp_el2.show_options["control_point_ids"] = False
    nur_el3.show_options["knots"] = False

    # Test Output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)

        # With a list
        splinepy.io.svg.export(
            tmpf,
            [bez_el0, rbz_el1, bsp_el2, nur_el3],
            indent=False,
            box_margins=0.7,
            font_family="sans-serif",
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/svg_noindent.svg"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

        # Without a list
        splinepy.io.svg.export(
            tmpf,
            bez_el0,
            rbz_el1,
            bsp_el2,
            nur_el3,
            indent=False,
            box_margins=0.7,
            font_family="sans-serif",
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/svg_noindent.svg"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

    # for python version > 3.9, test indented version
    if int(python_version.split(".")[1]) >= 9:
        with tempfile.TemporaryDirectory() as tmpd:
            tmpf = to_tmpf(tmpd)
            # With a list and also more options
            splinepy.io.svg.export(
                tmpf,
                [bez_el0, rbz_el1, bsp_el2, nur_el3],
                indent=True,
                box_margins=0.5,
                # Set some show_options
                c="blue",
                alpha=0.2,
                control_point_r=0.1,
                knots=False,
                control_points=True,
                control_mesh_c="purple",
                control_point_c="black",
                control_mesh_lw=0.05,
                font_family="sans-serif",
            )

            with open(tmpf) as tmp_read, open(
                os.path.dirname(__file__) + "/../data/svg_indent.svg"
            ) as base_file:
                assert are_stripped_lines_same(
                    base_file.readlines(), tmp_read.readlines(), True
                )

    # Check some spline_color and arrows
    askew_spline2D = splinepy.Bezier(
        degrees=[2, 2],
        control_points=[
            [0.0, 0.0],
            [1.0, 0.5],
            [2.0, 0.2],
            [0.5, 1.5],
            [1.0, 1.5],
            [1.5, 1.5],
            [0.0, 3.0],
            [1.0, 2.5],
            [2.0, 3.0],
        ],
    )

    def color_function(data, on):  # noqa ARG001
        """Some random function for colors"""
        return np.hstack(
            (
                np.cos(2 * np.pi * on[:, 0]).reshape(-1, 1),
                np.sin(2 * np.pi * on[:, 1]).reshape(-1, 1),
            )
        )

    plot_func_data = splinepy.SplineDataAdaptor(
        askew_spline2D, function=color_function
    )
    askew_spline2D.spline_data["field_function"] = plot_func_data
    askew_spline2D.show_options["data"] = "field_function"
    askew_spline2D.show_options["knot_c"] = "black"
    askew_spline2D.show_options["control_points"] = False
    askew_spline2D.show_options["resolutions"] = 100
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)

        # Without a list
        splinepy.io.svg.export(
            tmpf,
            askew_spline2D,
            indent=False,
            box_margins=0.7,
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/svg_data_field.svg"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

    # Plot arrows
    askew_spline2D.show_options["arrow_data_scale"] = 0.15
    askew_spline2D.show_options.pop("data")
    askew_spline2D.show_options["arrow_data"] = "field_function"
    askew_spline2D.show_options["knot_lw"] = 0.4
    askew_spline2D.show_options["c"] = (100, 100, 100)
    askew_spline2D.show_options["arrow_data_on"] = (
        splinepy.utils.data.cartesian_product(
            [np.linspace(0, 1, 20) for _ in range(2)]
        )
    )
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)

        # Without a list
        splinepy.io.svg.export(
            tmpf,
            askew_spline2D,
            indent=False,
            box_margins=0.7,
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/svg_arrow_data.svg"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

    # One last example minimal to reuse file for documentation

    spline = splinepy.Bezier(
        degrees=[2, 2],
        control_points=[
            [1, 0],
            [3, 1],
            [5, 0],  # First row
            [0, 2],
            [2, 3],
            [4, 2],  # Second row
            [1, 4],
            [3, 5],
            [5, 4],  # Third row
        ],
    )
    spline.show_options["c"] = (157, 157, 156)
    spline.show_options["knot_c"] = "red"
    spline.show_options["control_mesh"] = True
    spline.show_options["control_mesh_lw"] = 0.05
    spline.show_options["control_point_c"] = (0, 102, 157)
    spline.show_options["control_mesh_c"] = (0, 102, 157)
    spline.show_options["control_point_ids"] = False
    spline.show_options["control_point_r"] = 0.2

    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.svg.export(
            tmpf, spline, indent=False, box_margins=0.2, background=None
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/svg_mini_example.svg"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )

    # Test Export of gus vertices
    vertices = gus.Vertices(spline.cps)
    vertices.vertex_data["field"] = np.linspace(0, 1, spline.cps.shape[0])

    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.svg.export(
            tmpf,
            vertices,
            indent=False,
            box_margins=0.4,
            background="black",
        )

        with open(tmpf) as tmp_read, open(
            os.path.dirname(__file__) + "/../data/svg_vertices.svg"
        ) as base_file:
            assert are_stripped_lines_same(
                base_file.readlines(), tmp_read.readlines(), True
            )



"""
npz io.

keys in raw files are:
  - control_points : (ncpts, dim) np.ndarray, float64
  - degrees : (para_dim,) np.ndarray, int32
  - knot_vectors : (1,) np.ndarray, str
  - whatami : (1,) np.ndarray, str
"""

import numpy as _np

from splinepy import io as _io
from splinepy import splinepy_core as _splinepy_core


def load(
    fname,
):
    """
    Read a list of splines in `.npz` form.

    Parameters
    -----------
    fname: str

    Returns
    --------
    list_spline: list of NURBS / BSpline / Bezier / Rational Bezier
    """
    loaded = _np.load(fname)

    # Initialize an empty list to store the splines
    list_of_spline_dicts = []

    # Loop through keys in loaded dictionary and find the ones starting with 'spline_'
    for i in range(loaded["number_of_splines"]):
        # Initialize an empty dictionary to store the properties of the spline
        dict_spline = {}

        # Add the common properties of all splines
        dict_spline["control_points"] = loaded[f"spline_{i}_control_points"]
        dict_spline["degrees"] = loaded[f"spline_{i}_degrees"]

        # Add the weights if the spline is rational
        if f"spline_{i}_weights" in loaded:
            dict_spline["weights"] = loaded[f"spline_{i}_weights"]

        # Add the knot vectors if the spline has them
        if f"spline_{i}_knot_vectors_0" in loaded:
            kvs = []
            for j in range(dict_spline["degrees"].size):
                kvs.append(loaded[f"spline_{i}_knot_vectors_{j}"])
            dict_spline["knot_vectors"] = kvs

        # Append dictionary of spline to the list
        list_of_spline_dicts.append(dict_spline)

    return _io.ioutils.dict_to_spline(list_of_spline_dicts)


def export(fname, list_of_splines):
    """
    Save a list of splines as `.npz`.

    Parameters
    -----------
    fname: str
    splines: list of NURBS / BSpline / Bezier / Rational Bezier

    Returns
    --------
    None
    """

    # Checking for proper type of input
    if not isinstance(list_of_splines, list):
        if isinstance(list_of_splines, _splinepy_core.PySpline):
            list_of_splines = [list_of_splines]
        else:
            raise TypeError("Only a list of splines or a spline can be saved.")

    # Initialize an empty dictionary to store the properties of each spline
    property_dicts = {}

    # Add number of splines in list to dict
    property_dicts["number_of_splines"] = len(list_of_splines)

    # Loop through the list of splines and add their properties to the dictionary
    for i, spline in enumerate(list_of_splines):
        # Use a prefix to distinguish different splines
        prefix = f"spline_{i}_"

        # Add the common properties of all splines
        property_dicts[f"{prefix}degrees"] = spline.degrees
        property_dicts[f"{prefix}control_points"] = spline.control_points

        # Add the weights if the spline is rational
        if spline.is_rational:
            property_dicts[f"{prefix}weights"] = spline.weights

        # Add the knot vectors if the spline has them
        if spline.has_knot_vectors:
            for j in range(spline.para_dim):
                property_dicts[f"{prefix}knot_vectors_{j}"] = (
                    spline.knot_vectors[j]
                )

    # Save the dictionary as `.npz`
    _np.savez(
        fname,
        **property_dicts,
    )



"""Test mapping of basis function derivatives"""

import numpy as np
import pytest

import splinepy


@pytest.fixture
def scaling3D():
    return splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0.0, 0.0, 0.0],
            [2.0, 0.0, 0.0],
            [0.0, 0.3, 0.0],
            [2.0, 0.3, 0.0],
            [0.0, 0.0, 1.5],
            [2.0, 0.0, 1.5],
            [0.0, 0.3, 1.5],
            [2.0, 0.3, 1.5],
        ],
    )


@pytest.fixture
def rotation_matrix():
    cc, ss = np.cos(0.17), np.sin(0.17)
    return np.array(((cc, -ss), (ss, cc)))


@pytest.fixture
def rotating2D(rotation_matrix):
    rotating2D = splinepy.Bezier(
        degrees=[1, 1],
        control_points=[
            [0, 0],
            [1, 0],
            [0, 1],
            [1, 1],
        ],
    )
    rotating2D.control_points = np.einsum(
        "ij,qj->qi", rotation_matrix, rotating2D.control_points
    )
    return rotating2D


@pytest.fixture
def askew_spline2D():
    return splinepy.BSpline(
        degrees=[2, 2],
        control_points=[
            [0.0, 0.0],
            [1.0, 0.5],
            [2.0, 0.2],
            [0.5, 1.5],
            [1.0, 1.5],
            [1.5, 1.5],
            [0.0, 3.0],
            [1.0, 2.5],
            [2.0, 3.0],
        ],
        knot_vectors=[[0, 0, 0, 1, 1, 1], [0, 0, 0, 1, 1, 1]],
    )


@pytest.fixture
def solution_field_mono3D():
    return splinepy.Bezier(
        degrees=[2, 1, 1], control_points=np.ones((12, 1)) * 0.2
    )


@pytest.fixture
def solution_field_rando(np_rng):
    return splinepy.Bezier(
        degrees=[2, 1],
        control_points=np_rng.random((6, 1)),
    )


@pytest.fixture
def solution_field_rando2D(np_rng):
    return splinepy.Bezier(
        degrees=[2, 1],
        control_points=np_rng.random((6, 2)),
    )


@pytest.fixture
def query_points2D(np_rng):
    return np_rng.random((13, 2))


@pytest.fixture
def query_points3D(np_rng):
    return np_rng.random((17, 3))


def test_cross_evaluation_of_different_implementations(
    solution_field_rando2D,
    askew_spline2D,
    query_points2D,
):
    """Divergence and Laplacian are implemented differently when gradients
    are called at the same time
    """

    # Test Basis functions first
    mapper = solution_field_rando2D.mapper(askew_spline2D)
    bf_results = mapper.basis_function_derivatives(
        query_points2D, gradient=True, hessian=True, laplacian=True
    )
    bf_gradient, support_gradient = mapper.basis_gradient_and_support(
        query_points2D
    )
    bf_hessian, support_hessian = mapper.basis_hessian_and_support(
        query_points2D
    )
    bf_laplace, support_laplacian = mapper.basis_laplacian_and_support(
        query_points2D
    )

    # Laplacian is computed differently depending on function call
    assert np.allclose(bf_results["laplacian"], bf_laplace)
    assert np.allclose(bf_results["support"], support_laplacian)

    # Check function calls
    assert np.allclose(bf_results["gradient"], bf_gradient)
    assert np.allclose(bf_results["support"], support_gradient)
    assert np.allclose(bf_results["hessian"], bf_hessian)
    assert np.allclose(bf_results["support"], support_hessian)

    # Try field values
    results = mapper.field_derivatives(
        query_points2D,
        gradient=True,
        divergence=True,
        hessian=True,
        laplacian=True,
        basis_function_values=True,
    )
    laplacian = mapper.laplacian(query_points2D)
    divergence = mapper.divergence(query_points2D)
    gradient = mapper.gradient(query_points2D)
    hessian = mapper.hessian(query_points2D)

    # Divergence and laplacian have different implementations
    assert np.allclose(results["laplacian"], laplacian)
    assert np.allclose(results["divergence"], divergence)

    # Check function calls
    assert np.allclose(results["hessian"], hessian)
    assert np.allclose(results["gradient"], gradient)

    # Check with previous test
    assert np.allclose(
        results["basis_function_values"]["gradient"],
        bf_results["gradient"],
    )


def test_check_assertions(
    askew_spline2D, query_points2D, solution_field_rando
):
    mapper = solution_field_rando.mapper(askew_spline2D)
    with pytest.raises(
        ValueError,
        match=r"Divergence can only be performed "
        r"on vector fields with para_dim = dim",
    ):
        mapper.divergence(query_points2D)


def test_first_order_derivatives_analytical(
    solution_field_rando,
    rotating2D,
    solution_field_mono3D,
    scaling3D,
    query_points2D,
    query_points3D,
    rotation_matrix,
):
    mapper2D = solution_field_rando.mapper(rotating2D)
    mapper3D = solution_field_mono3D.mapper(scaling3D)
    bf_gradient, support = mapper2D.basis_gradient_and_support(query_points2D)
    (
        bf_reference0,
        supportb,
    ) = solution_field_rando.basis_derivative_and_support(
        query_points2D, [1, 0]
    )
    (
        bf_reference1,
        supportb,
    ) = solution_field_rando.basis_derivative_and_support(
        query_points2D, [0, 1]
    )
    bf_reference = np.dstack((bf_reference0, bf_reference1))

    assert np.allclose(support, supportb)

    # Rotate bf_reference
    bf_reference = np.einsum("ij,qsj->qsi", rotation_matrix, bf_reference)

    assert np.allclose(bf_gradient, bf_reference)

    bf_gradient, support = mapper3D.basis_gradient_and_support(query_points3D)
    (
        bf_reference0,
        supportb,
    ) = solution_field_mono3D.basis_derivative_and_support(
        query_points3D, [1, 0, 0]
    )
    (
        bf_reference1,
        supportb,
    ) = solution_field_mono3D.basis_derivative_and_support(
        query_points3D, [0, 1, 0]
    )
    (
        bf_reference2,
        supportb,
    ) = solution_field_mono3D.basis_derivative_and_support(
        query_points3D, [0, 0, 1]
    )
    bf_reference = np.dstack((bf_reference0, bf_reference1, bf_reference2))
    bf_reference = np.einsum(
        "qsi,i->qsi",
        bf_reference,
        1 / np.array([2.0, 0.3, 1.5]),
    )
    assert np.allclose(bf_gradient, bf_reference)


def test_second_order_analytical(
    solution_field_rando, rotating2D, query_points2D, rotation_matrix
):
    mapper2D = solution_field_rando.mapper(rotating2D)
    bf_hessian, support = mapper2D.basis_hessian_and_support(query_points2D)
    bf_reference = np.zeros(
        (
            query_points2D.shape[0],
            np.prod(solution_field_rando.degrees + 1),
            2,
            2,
        )
    )

    (
        bf_reference[:, :, 0, 0],
        supportb,
    ) = solution_field_rando.basis_derivative_and_support(
        query_points2D, [2, 0]
    )
    (
        bf_reference[:, :, 1, 0],
        supportb,
    ) = solution_field_rando.basis_derivative_and_support(
        query_points2D, [1, 1]
    )
    (
        bf_reference[:, :, 1, 1],
        supportc,
    ) = solution_field_rando.basis_derivative_and_support(
        query_points2D, [0, 2]
    )
    bf_reference[:, :, 0, 1] = bf_reference[:, :, 1, 0]

    assert np.allclose(support, supportb) and np.allclose(support, supportc)

    # Rotate bf_reference
    bf_reference = np.einsum(
        "ij,qsjk,kl->qsil",
        rotation_matrix,
        bf_reference,
        rotation_matrix.T,
    )

    assert np.allclose(bf_hessian, bf_reference)


def test_second_order_fd(np_rng, askew_spline2D, solution_field_rando2D):
    "Use proximity to get points on askew geometry and approcimate hessian"
    mapper = solution_field_rando2D.mapper(askew_spline2D)
    center_point_reference = np_rng.random((1, 2)) * 0.8 + 0.1
    dx = 1e-4

    # Analytical solution
    bf_references = mapper.basis_function_derivatives(
        queries=center_point_reference,
        gradient=True,
        hessian=True,
        laplacian=True,
    )
    references = mapper.field_derivatives(
        queries=center_point_reference,
        gradient=True,
        hessian=True,
        laplacian=True,
        divergence=True,
    )

    # Compute aux values for FD
    center_point = askew_spline2D.evaluate(center_point_reference)
    center_point = np.repeat(center_point, 9, axis=0)
    center_point += np.array(
        [
            [-dx, -dx],
            [0, -2 * dx],
            [dx, -dx],
            [-2 * dx, 0],
            [0, 0],
            [2 * dx, 0],
            [-dx, dx],
            [0, 2 * dx],
            [dx, dx],
        ]
    )
    # Approximate points in the physical domain
    center_point_parametric = askew_spline2D.proximities(
        center_point,
        initial_guess_sample_resolutions=[10, 10],
        tolerance=1e-12,
    )
    bfv, support = solution_field_rando2D.basis_and_support(
        center_point_parametric
    )
    values = solution_field_rando2D.evaluate(center_point_parametric)

    assert np.allclose(
        center_point_parametric[4, :],
        center_point_reference,
        atol=1e-8,
    )

    # Use FD to approximate
    # Gradient of basis function
    bf_gradient = np.zeros(bf_references["gradient"].shape)
    bf_gradient[:, :, 0] = (bfv[5, :] - bfv[3, :]) / (4 * dx)
    bf_gradient[:, :, 1] = (bfv[7, :] - bfv[1, :]) / (4 * dx)

    assert np.allclose(bf_gradient, bf_references["gradient"], atol=1e-4)

    # Hessians of basis function
    bf_hessian = np.zeros(bf_references["hessian"].shape)
    bf_hessian[:, :, 0, 0] = (bfv[5, :] + bfv[3, :] - 2 * bfv[4, :]) / (
        4 * dx * dx
    )
    bf_hessian[:, :, 1, 1] = (bfv[1, :] + bfv[7, :] - 2 * bfv[4, :]) / (
        4 * dx * dx
    )
    bf_hessian[:, :, 1, 0] = (
        bfv[8, :] - bfv[6, :] - bfv[2, :] + bfv[0, :]
    ) / (4 * dx * dx)
    bf_hessian[:, :, 0, 1] = bf_hessian[:, :, 1, 0]

    assert np.allclose(bf_hessian, bf_references["hessian"], atol=1e-4)

    # Gradient
    gradient = np.zeros(references["gradient"].shape)
    gradient[:, :, 0] = (values[5, :] - values[3, :]) / (4 * dx)
    gradient[:, :, 1] = (values[7, :] - values[1, :]) / (4 * dx)

    assert np.allclose(gradient, references["gradient"], atol=dx)

    # Hessians
    hessian = np.zeros(references["hessian"].shape)
    hessian[:, :, 0, 0] = (values[5, :] + values[3, :] - 2 * values[4, :]) / (
        4 * dx * dx
    )
    hessian[:, :, 1, 1] = (values[1, :] + values[7, :] - 2 * values[4, :]) / (
        4 * dx * dx
    )
    hessian[:, :, 1, 0] = (
        values[8, :] - values[6, :] - values[2, :] + values[0, :]
    ) / (4 * dx * dx)
    hessian[:, :, 0, 1] = hessian[:, :, 1, 0]

    assert np.allclose(hessian, references["hessian"], atol=dx)

    # Check reduced values
    laplacian = (
        references["hessian"][:, :, 0, 0] + references["hessian"][:, :, 1, 1]
    )
    divergence = (
        references["gradient"][:, 0, 0] + references["gradient"][:, 1, 1]
    )

    assert np.allclose(laplacian, references["laplacian"])
    assert np.allclose(divergence, references["divergence"])



import numpy as _np

from splinepy.bezier import Bezier as _Bezier
from splinepy.microstructure.tiles.tile_base import TileBase as _TileBase


class Chi(_TileBase):
    """Greek letter Chi resembling 2D curve tile.

    .. raw:: html

        <p><a href="../_static/Chi.html">Fullscreen</a>.</p>
        <embed type="text/html" width="100%" height="400" src="../_static/Chi.html" />

    """  # noqa: E501

    _dim = 2
    _para_dim = 1
    _evaluation_points = _np.array([[0.5, 0.5]])
    _n_info_per_eval_point = 1

    def create_tile(
        self,
        parameters=None,
        parameter_sensitivities=None,
        **kwargs,  # noqa ARG002
    ):
        """Create a microtile based on the parameter that describes a
        drilled cross around the center

        Parameters
        ----------
        parameters : np.array
          only first entry is used, defines the angle of the branches in
          radian, where zero results in a cross
        parameter_sensitivities: np.ndarray
          correlates with the drill angle

        Returns
        -------
        microtile_list : list(splines)
        """

        # set to default if nothing is given
        if parameters is None:
            self._logd(
                "Tile request is not parametrized, setting default Pi/8"
            )
            parameters = _np.array([[_np.pi / 8]])
        else:
            if not (
                _np.all(parameters >= -_np.pi * 0.5)
                and _np.all(parameters < _np.pi * 0.5)
            ):
                raise ValueError("The parameter must be in -Pi/2 and Pi/2")
            pass
        self.check_params(parameters)

        # Check if user requests derivative splines
        if self.check_param_derivatives(parameter_sensitivities):
            n_derivatives = parameter_sensitivities.shape[2]
            derivatives = []
        else:
            n_derivatives = 0
            derivatives = None

        splines = []
        for i_derivative in range(n_derivatives + 1):
            # Constant auxiliary values
            if i_derivative == 0:
                alpha = parameters[0, 0] + _np.pi / 4
                v_one_half = 0.5
                v_zero = 0.0
                r = _np.sqrt(0.125)
                s = r * _np.sin(alpha)
                c = r * _np.cos(alpha)
            else:
                alpha = parameter_sensitivities[0, 0, i_derivative - 1]
                v_one_half = 0.0
                v_zero = 0.0
                r = _np.sqrt(0.125)
                s = r * _np.cos(alpha)
                c = -r * _np.sin(alpha)

            # Init return value
            spline_list = []
            # 1
            spline_list.append(
                _Bezier(
                    degrees=[2],
                    control_points=_np.array(
                        [
                            [v_zero, v_zero],
                            [-s, -c],
                            [-v_one_half, -v_one_half],
                        ]
                    )
                    + v_one_half,
                )
            )
            # 2
            spline_list.append(
                _Bezier(
                    degrees=[2],
                    control_points=_np.array(
                        [
                            [v_zero, v_zero],
                            [c, -s],
                            [v_one_half, -v_one_half],
                        ]
                    )
                    + v_one_half,
                )
            )
            # 3
            spline_list.append(
                _Bezier(
                    degrees=[2],
                    control_points=_np.array(
                        [
                            [v_zero, v_zero],
                            [s, c],
                            [v_one_half, v_one_half],
                        ]
                    )
                    + v_one_half,
                )
            )
            # 4
            spline_list.append(
                _Bezier(
                    degrees=[2],
                    control_points=_np.array(
                        [
                            [v_zero, v_zero],
                            [-c, s],
                            [-v_one_half, v_one_half],
                        ]
                    )
                    + v_one_half,
                )
            )
            # Pass to output
            if i_derivative == 0:
                splines = spline_list.copy()
            else:
                derivatives.append(spline_list)

        # Return results
        return (splines, derivatives)



from splinepy.bspline import BSplineBase as _BSplineBase


class NURBS(_BSplineBase):
    r"""
    Non-Uniform Rational B-Spline.

    Passes all arguments to :code:`super.__init__()`,
    see :class:`.BSplineBase`.

    NURBS are an extension of B-Splines overcoming their drawback of being
    unable to represent circular shapes. This is achieved by the introduction
    of a weighting function
    :math:`W:\mathbb{R}^{N_{param}}\to\mathbb{R}^{+}_{*}`, defined as (for a
    one-dimensional parameter space, i.e., :math:`N_{param}=1`)

    .. math::
            W(u) = \sum_{i=1}^{l} N^{i;p}(u) w^i

    with scalar weights :math:`w^i\in\mathbb{R}^{+}_{*}`. Note, that the same
    basis functions are used for both the weighting function and projection
    space. Consequently, for the two-dimensional case (:math:`N_{param}=2`) we
    have

    .. math::
            W(u,v) = \sum_{i=1}^{l} \sum_{j=1}^{m} N^{i;p}(u) N^{j;q}(v)
                w^{i,j}

    and for the three-dimensional case (:math:`N_{param}=3`)

    .. math::
            W(u,v,w) = \sum_{i=1}^{l} \sum_{j=1}^{m} \sum_{k=1}^{n}
                N^{i;p}(u) N^{j;q}(v) N^{k;r}(w) w^{i,j,k}

    We can now construct different spline embeddings similar to the
    description in :class:`.BSplineBase`, only including the additional
    weighting:

    .. note::
        For simplicity, we restrict ourselves to the three most common types
        of splines in the following, namely curves, surfaces and volumes,
        although :code:`splinepy` also supports higher dimensions, see
        the documentation of :class:`.Spline` for more information.

    1. A NURBS of degree :math:`p` with control points
    :math:`P^i\in\mathbb{R}^{N_{phys}}`, weights
    :math:`w^i\in\mathbb{R}^{+}_{*}`, and a one-dimensional parameter space
    (:math:`N_{param}=1`) corresponds to a line, embedded into the physical
    space:

    .. math::
            C(u) = \sum_{i=1}^{l} R^{i;p}(u) P^i

    with the modified (rational) basis functions

    .. math::
            R^{i;p}(u) = \frac{N^{i;p}(u)w^i}{W(u)}

    2. A NURBS of degrees :math:`p,q` with control points
    :math:`P^{i,j}\in\mathbb{R}^{N_{phys}}`, weights
    :math:`w^{i,j}\in\mathbb{R}^{+}_{*}`, and a two-dimensional parameter
    space (:math:`N_{param}=2`) corresponds to a surface, embedded into the
    physical space:

    .. math::
            S(u,v) = \sum_{i=1}^{l} \sum_{j=1}^{m} R^{i,j;p,q}(u,v) P^{i,j}

    with the modified (rational) basis functions

    .. math::
            R^{i,j;p,q}(u,v) =
                \frac{\tilde{N}^{i,j;p,q}(u,v) w^{i,j}}{W(u,v)}

    3. A NURBS of degrees :math:`p,q,r` with control points
    :math:`P^{i,j,k}\in\mathbb{R}^{N_{phys}}`, weights
    :math:`w^{i,j,k}\in\mathbb{R}^{+}_{*}`, and a three-dimensional parameter
    space (:math:`N_{param}=3`) corresponds to a volume, embedded into the
    physical space:

    .. math::
            V(u,v,w) = \sum_{i=1}^{l} \sum_{j=1}^{m} \sum_{k=1}^{n}
                R^{i,j,k;p,q,r}(u,v,w) P^{i,j,k}

    with the modified (rational) basis functions

    .. math::
            R^{i,j,k;p,q,r}(u,v,w) =
                \frac{
                    \tilde{N}^{i,j,k;p,q,r}(u,v,w) w^{i,j,k}
                }{
                    W(u,v,w)
                }

    Higher-dimensional instances are constructed accordingly.

    **Usage**:

    .. code-block:: python

        # NURBS curve (circle)
        nurbs_curve = splinepy.NURBS(
            degrees=[2],
            knot_vectors=[
                0.0,
                0.0,
                0.0,
                0.25,
                0.25,
                0.5,
                0.5,
                0.75,
                0.75,
                1.0,
                1.0,
                1.0,
            ],
            control_points=[
                [1.0, 0.0],
                [1.0, 1.0],
                [0.0, 1.0],
                [-1.0, 1.0],
                [-1.0, 0.0],
                [-1.0, -1.0],
                [0.0, -1.0],
                [1.0, -1.0],
                [1.0, 0.0],
            ],
        )

    Parameters
    -----------
    degrees: (para_dim,) list-like
    knot_vectors: (para_dim, n) list
    control_points: (m, dim) list-like
    weights: (m,) list-like

    Returns
    --------
    None
    """

    __slots__ = ()

    def __init__(
        self,
        degrees=None,
        knot_vectors=None,
        control_points=None,
        weights=None,
        spline=None,
    ):
        super().__init__(
            spline=spline,
            degrees=degrees,
            knot_vectors=knot_vectors,
            control_points=control_points,
            weights=weights,
        )

    @property
    def nurbs(self):
        return self.copy(saved_data=False)



import gustaf as _gus
import numpy as _np
from gustaf import Vertices as _Vertices
from gustaf.helpers import options as _options
from gustaf.utils.arr import enforce_len as _enforce_len

from splinepy import settings as _settings
from splinepy.utils import log as _log

_vedo_spline_common_options = (
    _options.Option("vedo", "knots", "Show spline's knots.", (bool,)),
    _options.Option("vedo", "knot_c", "Knot color.", (str, tuple, list, int)),
    _options.Option(
        "vedo",
        "knot_lw",
        "Line width of knots. Number of pixels. "
        "Applicable to para_dim > 1.",
        (float, float),
    ),
    _options.Option(
        "vedo",
        "knot_alpha",
        "Transparency of knots in range [0, 1]. Applicable to para_dim > 1",
        (float, int),
    ),
    _options.Option(
        "vedo",
        "control_points",
        "Show spline's control points."
        "Options propagates to control mesh, unless it is specified.",
        (bool,),
    ),
    _options.Option(
        "vedo",
        "control_point_r",
        "Control point radius",
        (float, int),
    ),
    _options.Option(
        "vedo",
        "control_point_c",
        "Color of control_points in {rgb, RGB, str of (hex, name), int}",
        (str, tuple, list, int),
    ),
    _options.Option(
        "vedo",
        "control_point_alpha",
        "Transparency of control points in range [0, 1].",
        (float, int),
    ),
    _options.Option(
        "vedo",
        "control_point_ids",
        "Show ids of control_points.",
        (bool,),
    ),
    _options.Option(
        "vedo",
        "control_mesh",
        "Show spline's control mesh.",
        (bool,),
    ),
    _options.Option(
        "vedo",
        "control_mesh_c",
        "Color of control_mesh in {rgb, RGB, str of (hex, name), int}",
        (str, tuple, list, int),
    ),
    _options.Option(
        "vedo",
        "control_mesh_lw",
        "Line width of control mesh. Number of pixels",
        (float, int),
    ),
    _options.Option(
        "vedo",
        "resolutions",
        "Sampling resolution for spline.",
        (int, list, tuple, _np.ndarray),
    ),
    # overwrites the option with a same name in vedo_common_options
    _options.Option(
        "vedo",
        "lw",
        "Line width 1D spline." "Applicable to para_dim > 1.",
        (float, int),
    ),
)


class SplineShowOption(_options.ShowOption):
    """
    Show options for splines.
    """

    __slots__ = ()

    # if we start to support more backends, most of this options should become
    # some sort of spline common.
    _valid_options = _options.make_valid_options(
        *_options.vedo_common_options,
        *_vedo_spline_common_options,
        _options.Option(
            "vedo",
            "fitting_queries",
            "Shows fitting queries if they are locally saved in splines.",
            (bool,),
        ),
        _options.Option(
            "vedo",
            "arrow_data_on",
            "Specify parametric coordinates to place arrow_data.",
            (list, tuple, _np.ndarray),
        ),
    )

    _helps = "Spline"

    def __init__(self, helpee):
        """
        Parameters
        ----------
        helpee: GustafSpline
        """
        self._helpee = helpee
        # checks if helpee inherits from GustafSpline
        if self._helps not in str(type(helpee).__mro__):
            raise TypeError(
                f"This show option if for {self._helps}.",
                f"Given helpee is {type(helpee)}.",
            )
        self._options = {}


class MultipatchShowOption(_options.ShowOption):
    """
    Show options for Multipatches.
    """

    __slots__ = ()

    # if we start to support more backends, most of this options should become
    # some sort of spline common.
    _valid_options = _options.make_valid_options(
        *_options.vedo_common_options,
        *_vedo_spline_common_options,
    )

    _helps = "Multipatch"

    def __init__(self, helpee):
        """
        Parameters
        ----------
        helpee: GustafSpline
        """
        self._helpee = helpee
        # checks if helpee inherits from GustafSpline
        if self._helps not in str(type(helpee).__mro__):
            raise TypeError(
                f"This show option if for {self._helps}.",
                f"Given helpee is {type(helpee)}.",
            )
        self._options = {}


def _sample_knots_1d(spline):
    """
    Samples knots for splines of para_dim=1.
    Assumes this is one of the showable splines.
    """

    def sample_uks(s):
        """samples unique knots from single patch spline"""
        return s.evaluate(_np.asanyarray(s.unique_knots[0]).reshape(-1, 1))

    # sample unique knots in physical space and create vertices
    # for multi-patch, we will call this function for its patches
    if spline.name.startswith("Multi"):
        evaluated_knots = _np.vstack([sample_uks(p) for p in spline.patches])
    else:
        evaluated_knots = sample_uks(spline)

    # create vertices - for multipatch, we could remove duplicating vertices
    # it gets annoying
    knots = _Vertices(evaluated_knots)

    # apply show options to show this vertices as 'x'
    knots.show_options["labels"] = ["x"] * len(knots.vertices)
    knots.show_options["label_options"] = {
        "justify": "center",
        "c": spline.show_options.get("knot_c", "green"),
    }

    return knots


def _sample_knots_2d(spline, res):
    """
    Samples knots for splines of para_dim=2.
    Assumes this is one of the showable splines.
    """
    # extract knot lines as edges
    knot_lines = spline.extract.edges(res, all_knots=True)

    # apply show options
    knot_lines.show_options["c"] = spline.show_options.get("knot_c", "black")
    knot_lines.show_options["lw"] = spline.show_options.get("knot_lw", 3)
    knot_lines.show_options["alpha"] = spline.show_options.get("knot_alpha", 1)

    return knot_lines


def _sample_knots(spline, res=None):
    """
    Samples knots for splines.
    Assumes this is one of the showable splines
    """
    # call sample functions for corresponding para_dim
    para_dim = spline.para_dim
    if para_dim > 1:
        if res is None:
            raise ValueError("sample resolution not given.")
        # 2d and 3d splines has same knot sampling routine
        return _sample_knots_2d(spline, res)
    else:
        return _sample_knots_1d(spline)


def _sample_spline(spline, res):
    """
    Sample splines.
    Assumes this is one of the showable splines

    Parameters
    ----------
    None

    Returns
    -------
    gus_dict: dict
      dict of sampled spline and knots.
    """
    # sampling spline is quite simple, so we don't split functions here
    para_dim = spline.para_dim
    if para_dim == 1:
        sp = spline.extract.edges(res)
        sp.show_options["lw"] = spline.show_options.get("lw", 8)
    elif para_dim > 1:
        # we don't return water tight meshes here
        # see (gustaf #120)
        sp = spline.extract.faces(res, watertight=False)

    return sp


def _process_scalar_field(spline, data, sampled_spline, res):
    """
    Assuming field data is defined, extracts field value and applies
    to sampled spline.
    """
    # get data name and try to sample scalar field
    sampled_spline_data = spline.spline_data.as_scalar(data, res, default=None)

    # apply sampled data as vertex data to sampled_spline
    if sampled_spline_data is not None:
        # transfer data
        sampled_spline.vertex_data[data] = sampled_spline_data

        # transfer options - maybe vectorized query?
        keys = (
            "vmin",
            "vmax",
            "scalarbar",
            "scalarbar3d",
            "cmap",
            "cmap_alpha",
            "cmap_n_colors",
        )
        spline.show_options.copy_valid_options(
            sampled_spline.show_options, keys
        )

        # mark that we want to see this data
        sampled_spline.show_options["data"] = data

    else:
        raise ValueError(f"No spline_data named ({data}) for {spline}.")


def _process_spline_color(spline, sampled_spline, res):
    """
    Processes color and scalar field.
    If both are defined, you will see the color of the scalar field.
    """
    # first, get defaults
    default_lighting = "glossy" if spline.dim > 2 else "off"
    default_color = "green" if spline.para_dim > 1 else "black"

    # apply basic color
    sampled_spline.show_options["c"] = spline.show_options.get(
        "c", default_color
    )

    # apply alpha and lighting
    sampled_spline.show_options["alpha"] = spline.show_options.get("alpha", 1)
    sampled_spline.show_options["lighting"] = spline.show_options.get(
        "lighting", default_lighting
    )

    # process scalar field if data is defined
    data = spline.show_options.get("data", None)
    if data is not None:
        _process_scalar_field(spline, data, sampled_spline, res)


def _sample_arrow_data(spline, adata, sampled_spline, res):
    """
    Process arrow_data. Also known as vector_data.
    """
    # early exit?
    if adata is None:
        return None

    # use getitem to retrieve adapted_adata. this will raise key error if
    # it doesn't exist
    adapted_adata = spline.spline_data[adata]

    # if location is specified, this will be a separate Vertices obj with
    # configured arrow_data. Multipatch might just return a multipatch,
    # so a careful step here.
    has_locations = getattr(adapted_adata, "has_locations", False)
    adata_on = "arrow_data_on" in spline.show_options.keys()  # noqa SIM118
    create_vertices = has_locations or adata_on

    # this case causes conflict of interest. raise
    if has_locations and adata_on:
        raise ValueError(
            f"arrow_data-({adata}) has fixed location, "
            "but and `arrow_data_on` is set.",
        )

    # two cases of arrow data sampling
    # 1 - same as sampled spline
    #      -> add arrow as vertex_data
    # 2 - at specified locations
    #      -> create vertices at specified locations
    #         then, add arrow as vertex_data to them
    if not create_vertices:
        # sample arrows and append to sampled spline.
        sampled_spline.vertex_data[adata] = spline.spline_data.as_arrow(
            adata, resolutions=res
        )

        # transfer options to sampled_spline
        keys = ("arrow_data_scale", "arrow_data_color", "arrow_data")
        spline.show_options.copy_valid_options(
            sampled_spline.show_options, keys
        )

    else:
        # prepare corresponding queries
        if getattr(adapted_adata, "has_locations", False):
            queries = adapted_adata.locations
            on = None
        else:
            queries = spline.show_options["arrow_data_on"]
            on = queries

        # bound /  dim check - only for splines
        if not spline.name.startswith("Multi"):
            bounds = spline.parametric_bounds
            if queries.shape[1] != len(bounds[0]):
                raise ValueError(
                    "Dimension mismatch: arrow_data locations-"
                    f"{queries.shape[1]} / para_dim-{spline.para_dim}."
                )
            # tolerance padding. may still cause issues in splinepy.
            # in that case, we will have to scale queries.
            lb_diff = queries.min(axis=0) - bounds[0] + _settings.TOLERANCE
            ub_diff = queries.max(axis=0) - bounds[1] - _settings.TOLERANCE
            if any(lb_diff < 0) or any(ub_diff > 0):
                raise ValueError(
                    f"Given locations for ({adata}) are outside the "
                    f"parametric bounds ({bounds})."
                )

        # get arrow
        adata_values = spline.spline_data.as_arrow(adata, on=on)

        # create vertices that can be shown as arrows
        loc_vertices = _Vertices(spline.evaluate(queries))
        loc_vertices.vertex_data[adata] = adata_values

        # transfer options
        keys = ("arrow_data_scale", "arrow_data_color", "arrow_data")
        spline.show_options.copy_valid_options(loc_vertices.show_options, keys)

        return loc_vertices


def _create_control_points(spline):
    """
    Creates control points.
    """
    # control points (vertices)
    cps = spline.extract.control_points()
    cps.show_options["c"] = spline.show_options.get("control_point_c", "red")
    cps.show_options["r"] = spline.show_options.get("control_point_r", 10)
    cps.show_options["alpha"] = spline.show_options.get(
        "control_point_alpha", 1.0
    )
    return cps


def _create_control_point_ids(spline):
    """
    Creates control point ids.
    """
    # a bit redundant, but nicely separable
    cp_ids = spline.extract.control_points()
    cp_ids.show_options["labels"] = _np.arange(len(cp_ids.vertices))
    cp_ids.show_options["label_options"] = {"font": "VTK"}

    return cp_ids


def _create_control_mesh(spline):
    """
    Creates control mesh.
    """
    # pure control mesh
    c_mesh = spline.extract.control_mesh()  # either edges or faces
    if spline.para_dim != 1:
        c_mesh = c_mesh.to_edges(unique=True)

    c_mesh.show_options["c"] = spline.show_options.get("control_mesh_c", "red")
    c_mesh.show_options["lw"] = spline.show_options.get("control_mesh_lw", 4)
    c_mesh.show_options["alpha"] = spline.show_options.get(
        "control_points_alpha", 0.8
    )

    return c_mesh


def _create_fitting_queries(spline):
    """
    Create fitting queries. This will raise if spline is not BSpline or
    fitting queries does not exist.
    """
    fqs = _Vertices(spline._fitting_queries)
    fqs.show_options["c"] = "blue"
    fqs.show_options["r"] = 10
    return fqs


def make_showable(spline):
    """
    Goes through common procedures for preparing showable splines.

    Parameters
    ----------
    None

    Returns
    -------
    gus_dict: dict
      dict of sampled spline as gustaf objects
    """
    # get sampling resolution
    default_res = 10 if spline.name.startswith("Multi") else 100
    res = spline.show_options.get(
        "resolutions", default_res
    )  # , spline.para_dim)

    # get spline and knots
    sampled_gus = {}
    sampled_spline = _sample_spline(spline, res)
    sampled_gus["spline"] = sampled_spline  # save to dict
    if spline.show_options.get("knots", True):
        sampled_gus["knots"] = _sample_knots(spline, res)

    # apply spline color - this includes scalar field
    _process_spline_color(spline, sampled_spline, res)

    # axis?
    sampled_spline.show_options["axes"] = spline.show_options.get(
        "axes", False
    )

    # process arrow data, this may return something if arrow_data_on is defined
    arrow_data = _sample_arrow_data(
        spline,
        spline.show_options.get("arrow_data", None),
        sampled_spline,
        _enforce_len(res, spline.para_dim),
    )
    if arrow_data is not None:
        sampled_gus["arrow_data"] = arrow_data

    # process control_points
    show_cps = spline.show_options.get("control_points", True)
    if show_cps:
        sampled_gus["control_points"] = _create_control_points(spline)

    # control_point_ids
    if spline.show_options.get("control_point_ids", show_cps):
        sampled_gus["control_point_ids"] = _create_control_point_ids(spline)

    # control_mesh
    if spline.show_options.get("control_mesh", show_cps):
        sampled_gus["control_mesh"] = _create_control_mesh(spline)

    # fitting queries
    if hasattr(spline, "_fitting_queries") and spline.show_options.get(
        "fitting_queries", True
    ):
        sampled_gus["fitting_queries"] = _create_fitting_queries(spline)

    return sampled_gus


def show(spline, **kwargs):
    """Shows splines with various options.

    They are excessively listed, so that it can be adjustable.

    Parameters
    -----------
    spline: Spline / Multipatch
    resolutions: int or (spline.para_dim,) array-like
    control_points: bool
    knots: bool
    fitting_queries: bool
    return_gustaf: bool
      Return dict of gustaf discrete objects, for example,
      {Vertices, Edges, Faces}, instead of opening a window
    return_showable: bool
      Return dict of showable objects.
    parametric_space: bool
      Only relevant for `vedo` backend.
    c: str
      Default is None. Black for curves, else green.
    alpha: float
    lighting: str
    control_point_ids: bool
    color_spline: str
    cmap: str

    Returns
    --------
    things_to_show: dict
      iff return_discrete==True, dict of gustaf objects that are showable.
      iff return_showable==True, dict of backend objects that are showable.
    """
    # Showing is only possible for following splines
    allowed_dim_combo = (
        (1, 2),
        (1, 3),
        (2, 2),
        (2, 3),
        (3, 3),
    )
    if (spline.para_dim, spline.dim) not in allowed_dim_combo:
        raise ValueError("Sorry, can't show given spline.")

    orig_show_options = None
    if kwargs:
        orig_show_options = spline.show_options
        spline._show_options = spline.__show_option__(spline)
        orig_show_options.copy_valid_options(spline.show_options)
        for key, value in kwargs.items():
            try:
                spline.show_options[key] = value
            except BaseException:
                _log.debug(
                    f"Skipping invalid option {key} for "
                    f"{spline.show_options._helps}."
                )
                continue

    # Prepare things to show dict.
    things_to_show = make_showable(spline)

    # set original options back
    if orig_show_options is not None:
        spline._show_options = orig_show_options

    para_space = kwargs.pop("parametric_space", False)
    if para_space:
        p_spline = spline.create.parametric_view()
        things_to_show["parametric_spline"] = p_spline

    if kwargs.get("return_gustaf", False):
        return things_to_show

    if kwargs.get("return_showable", False):
        return {key: value.showable() for key, value in things_to_show.items()}

    if para_space:
        things_to_show.pop("parametric_spline")

        return _gus.show(
            ["Parametric Space", p_spline],
            ["Physical Space", *things_to_show.values()],
            **kwargs,
        )

    return _gus.show(things_to_show, **kwargs)



import tempfile

import splinepy


def test_irit_export_import(
    bspline_2p2d, nurbs_2p2d, bezier_3p3d, to_tmpf, are_splines_equal
):
    """Test import export routine for irit files"""
    # Create a line
    rational = nurbs_2p2d.extract.spline(splitting_plane=0, interval=0.3)

    # Make it more tricky
    bspline_2p2d.elevate_degrees([0, 1, 0])
    nurbs_2p2d.elevate_degrees([1])
    bezier_3p3d.elevate_degrees([0, 1, 2])
    nurbs_2p2d.insert_knots(1, [0.5])
    bspline_2p2d.insert_knots(1, [0.5])

    list_of_splines = [bezier_3p3d, nurbs_2p2d, bspline_2p2d, rational]

    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.irit.export(tmpf, list_of_splines)
        list_of_splines_loaded = splinepy.io.irit.load(tmpf)
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(list_of_splines, list_of_splines_loaded)
        )



import tempfile
from sys import version as python_version

import splinepy

cats_no_indent_export = [
    '<SplineList SplineType="1" NumberOfSplines="4"><!--generated by splinepy'
    ' https://github.com/tataratat/splinepy--><SplineEntry splDim="'
    '2" spaceDim="2" numOfCntrlPntVars="2" numCntrlPnts="6" numOfEleVars="0" c'
    'losed="0 0"><cntrlPntVarNames>x y</cntrlPntVarNames><cntrlPntVars>0.0 0.0'
    " 0.5 0.0 1.0 0.0 0.0 1.0 0.5 1.0 1.0 1.0</cntrlPntVars><deg>2 1</deg><knt"
    "Vecs><kntVec>0.0 0.0 0.0 1.0 1.0 1.0</kntVec><kntVec>0.0 0.0 1.0 1.0</"
    'kntVec></kntVecs></SplineEntry><SplineEntry splDim="2" spaceDim="2" numO'
    'fCntrlPntVars="2" numCntrlPnts="4" numOfEleVars="0" closed="0 0"><cntrlPn'
    "tVarNames>x y</cntrlPntVarNames><cntrlPntVars>1.0 0.0 2.0 0.0 1.0 1.0 2.0"
    " 1.0</cntrlPntVars><deg>1 1</deg><kntVecs><kntVec>0.0 0.0 1.0 1.0</knotV"
    "ec><kntVec>0.0 0.0 1.0 1.0</kntVec></kntVecs><wght>1.0 1.0 1.0 1.0</wgh"
    't></SplineEntry><SplineEntry splDim="2" spaceDim="2" numOfCntrlPntVars="2'
    '" numCntrlPnts="9" numOfEleVars="0" closed="0 0"><cntrlPntVarNames>x y</c'
    "ntrlPntVarNames><cntrlPntVars>0.0 1.0 0.5 1.0 1.0 1.0 0.0 1.5 0.5 1.5 1.0"
    " 1.5 0.0 2.0 0.5 2.0 1.0 2.0</cntrlPntVars><deg>2 1</deg><kntVecs><knotVe"
    "c>0.0 0.0 0.0 1.0 1.0 1.0</kntVec><kntVec>0.0 0.0 0.5 1.0 1.0</kntVec>"
    '</kntVecs></SplineEntry><SplineEntry splDim="2" spaceDim="2" numOfCntrlPn'
    'tVars="2" numCntrlPnts="6" numOfEleVars="0" closed="0 0"><cntrlPntVarName'
    "s>x y</cntrlPntVarNames><cntrlPntVars>1.0 1.0 2.0 1.0 1.0 1.5 2.0 1.5 1.0"
    " 2.0 2.0 2.0</cntrlPntVars><deg>1 1</deg><kntVecs><kntVec>0.0 0.0 1.0 1."
    "0</kntVec><kntVec>0.0 0.0 0.5 1.0 1.0</kntVec></kntVecs><wght>1.0 1.0 "
    "1.0 1.0 1.0 1.0</wght></SplineEntry></SplineList>"
]

cats_indent_export = [
    '<SplineList SplineType="1" NumberOfSplines="4">\n',
    "  <!--generated by splinepy https://github.com/tataratat/splinepy-->\n",
    '  <SplineEntry splDim="2" spaceDim="2" numOfCntrlPntVars="2" '
    'numCntrlPnts="6" numOfEleVars="0" closed="0 0">\n',
    "    <cntrlPntVarNames>x y</cntrlPntVarNames>\n",
    "    <cntrlPntVars>0.0 0.0\n",
    "0.5 0.0\n",
    "1.0 0.0\n",
    "0.0 1.0\n",
    "0.5 1.0\n",
    "1.0 1.0</cntrlPntVars>\n",
    "    <deg>2\n",
    "1</deg>\n",
    "    <kntVecs>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "0.0\n",
    "1.0\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "    </kntVecs>\n",
    "  </SplineEntry>\n",
    '  <SplineEntry splDim="2" spaceDim="2" numOfCntrlPntVars="2" '
    'numCntrlPnts="4" numOfEleVars="0" closed="0 0">\n',
    "    <cntrlPntVarNames>x y</cntrlPntVarNames>\n",
    "    <cntrlPntVars>1.0 0.0\n",
    "2.0 0.0\n",
    "1.0 1.0\n",
    "2.0 1.0</cntrlPntVars>\n",
    "    <deg>1\n",
    "1</deg>\n",
    "    <kntVecs>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "    </kntVecs>\n",
    "    <wght>1.0\n",
    "1.0\n",
    "1.0\n",
    "1.0</wght>\n",
    "  </SplineEntry>\n",
    '  <SplineEntry splDim="2" spaceDim="2" numOfCntrlPntVars="2" '
    'numCntrlPnts="9" numOfEleVars="0" closed="0 0">\n',
    "    <cntrlPntVarNames>x y</cntrlPntVarNames>\n",
    "    <cntrlPntVars>0.0 1.0\n",
    "0.5 1.0\n",
    "1.0 1.0\n",
    "0.0 1.5\n",
    "0.5 1.5\n",
    "1.0 1.5\n",
    "0.0 2.0\n",
    "0.5 2.0\n",
    "1.0 2.0</cntrlPntVars>\n",
    "    <deg>2\n",
    "1</deg>\n",
    "    <kntVecs>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "0.0\n",
    "1.0\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "0.5\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "    </kntVecs>\n",
    "  </SplineEntry>\n",
    '  <SplineEntry splDim="2" spaceDim="2" numOfCntrlPntVars="2" '
    'numCntrlPnts="6" numOfEleVars="0" closed="0 0">\n',
    "    <cntrlPntVarNames>x y</cntrlPntVarNames>\n",
    "    <cntrlPntVars>1.0 1.0\n",
    "2.0 1.0\n",
    "1.0 1.5\n",
    "2.0 1.5\n",
    "1.0 2.0\n",
    "2.0 2.0</cntrlPntVars>\n",
    "    <deg>1\n",
    "1</deg>\n",
    "    <kntVecs>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "      <kntVec>0.0\n",
    "0.0\n",
    "0.5\n",
    "1.0\n",
    "1.0</kntVec>\n",
    "    </kntVecs>\n",
    "    <wght>1.0\n",
    "1.0\n",
    "1.0\n",
    "1.0\n",
    "1.0\n",
    "1.0</wght>\n",
    "  </SplineEntry>\n",
    "</SplineList>",
]


def test_cats_export(to_tmpf, are_items_same, are_stripped_lines_same):
    """Test cats export routine"""
    ###########
    # 2D Mesh #
    ###########
    # Define some splines
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 1], [1, 1], [0, 2], [1, 2]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bez_el0.elevate_degrees(0)
    bsp_el2.elevate_degrees(0)

    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Init multipatch
    multipatch = splinepy.Multipatch(
        splines=[bez_el0, rbz_el1, bsp_el2, nur_el3]
    )

    # Test Output
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.cats.export(
            tmpf, spline_list=multipatch, indent=False, make_rational=False
        )

        with open(tmpf) as tmp_read:
            assert are_stripped_lines_same(
                cats_no_indent_export, tmp_read.readlines(), True
            )

    # for python version > 3.9, test indented version
    if int(python_version.split(".")[1]) >= 9:
        with tempfile.TemporaryDirectory() as tmpd:
            tmpf = to_tmpf(tmpd)
            splinepy.io.cats.export(
                tmpf, spline_list=multipatch, indent=True, make_rational=False
            )

            with open(tmpf) as tmp_read:
                assert are_items_same(cats_indent_export, tmp_read.readlines())


def test_cats_import(to_tmpf, are_splines_equal):
    """Test cats import routine"""
    # Define some splines
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1, 1],
        control_points=[
            [0, 1, 0],
            [1, 1, 0],
            [0, 2, 0],
            [1, 2, 0],
            [0, 1, 1],
            [1, 1, 1],
            [0, 2, 1],
            [1, 2, 1],
        ],
        knot_vectors=[[0, 0, 1, 1]] * 3,
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1, 1],
        control_points=[
            [1, 1, 0],
            [2, 1, 0],
            [1, 2, 0],
            [2, 2, 0],
            [1, 1, 1],
            [2, 1, 1],
            [1, 2, 1],
            [2, 2, 1],
        ],
        weights=[1] * 8,
        knot_vectors=[[0, 0, 1, 1]] * 3,
    )

    # Make it more tricky
    bsp_el2.elevate_degrees(0)
    bsp_el2.elevate_degrees(1)
    nur_el3.elevate_degrees(0)
    nur_el3.elevate_degrees(1)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Test Output against input
    multipatch_geometry = [bsp_el2, nur_el3]
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.cats.export(
            tmpf,
            spline_list=multipatch_geometry,
            indent=False,
            make_rational=False,
        )
        multipatch_geometry_loaded = splinepy.io.cats.load(tmpf)
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(
                multipatch_geometry,
                multipatch_geometry_loaded,
            )
        )

    # Make all splines rational
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.cats.export(
            tmpf,
            spline_list=multipatch_geometry,
            indent=False,
        )
        multipatch_geometry_loaded = splinepy.io.cats.load(tmpf)
        assert all(
            are_splines_equal(a.nurbs, b)
            for a, b in zip(
                multipatch_geometry,
                multipatch_geometry_loaded,
            )
        )



import tempfile

import splinepy


def test_json_import(to_tmpf, are_splines_equal):
    """
    Test json export-import routine
    """
    bez_el0 = splinepy.Bezier(
        degrees=[1, 1], control_points=[[0, 0], [1, 0], [0, 1], [1, 1]]
    )
    rbz_el1 = splinepy.RationalBezier(
        degrees=[1, 1],
        control_points=[[1, 0], [2, 0], [1, 1], [2, 1]],
        weights=[1, 1, 1, 1],
    )
    bsp_el2 = splinepy.BSpline(
        degrees=[1, 1],
        control_points=[[0, 1], [1, 1], [0, 2], [1, 2]],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )
    nur_el3 = splinepy.NURBS(
        degrees=[1, 1],
        control_points=[[1, 1], [2, 1], [1, 2], [2, 2]],
        weights=[1, 1, 1, 1],
        knot_vectors=[[0, 0, 1, 1], [0, 0, 1, 1]],
    )

    # Make it more tricky
    bsp_el2.elevate_degrees(0)
    bsp_el2.elevate_degrees(1)
    nur_el3.elevate_degrees(0)
    nur_el3.elevate_degrees(1)
    bsp_el2.insert_knots(1, [0.5])
    nur_el3.insert_knots(1, [0.5])

    # Test Output against input using base64 encoding
    list_of_splines = [bez_el0, rbz_el1, bsp_el2, nur_el3]
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.json.export(tmpf, list_of_splines, base64encoding=True)
        list_of_splines_loaded = splinepy.io.json.load(tmpf)
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(list_of_splines, list_of_splines_loaded)
        )

    # Test Import export with non base64 encoding
    with tempfile.TemporaryDirectory() as tmpd:
        tmpf = to_tmpf(tmpd)
        splinepy.io.json.export(tmpf, list_of_splines, base64encoding=False)
        list_of_splines_loaded = splinepy.io.json.load(tmpf)
        assert all(
            are_splines_equal(a, b)
            for a, b in zip(list_of_splines, list_of_splines_loaded)
        )



import numpy as np
import pytest

import splinepy


@pytest.fixture
def self_referencing_spline():
    # Spline that points to itself
    return splinepy.Bezier(
        degrees=[3, 1],
        control_points=[
            [0, 0],
            [2, 3],
            [-2, 3],
            [0, 0],
            [0, 1],
            [1, 2],
            [-1, 2],
            [0, 1],
        ],
    )


@pytest.fixture
def rotated_l_splines():
    # List of splinewith with this topology
    #
    #  o ---- o ---- o
    #  |   0  |   1  |
    #  o ---- o ---- o
    #  |   2  |
    #  o ---- o
    #
    rect_arc_1 = splinepy.Bezier(
        [2, 1], [[0, 0], [1, 0], [2, 0], [0, 1], [1, 2], [2, 1]]
    )
    rect_arc_2 = splinepy.Bezier([1, 1], [[2, 0], [3, 0], [2, 1], [3, 1]])
    rect_arc_3 = splinepy.Bezier(
        [2, 1], [[0, -1], [1, -2], [2, -1], [0, 0], [1, 0], [2, 0]]
    )
    return [
        rect_arc_1,
        rect_arc_2,
        rect_arc_3,
    ]


def test_interfaces(self_referencing_spline, rotated_l_splines):
    """ """
    # init multipatch with multiple splines
    multipatch = splinepy.Multipatch()
    multipatch.patches = rotated_l_splines

    # init multipatch with single spline
    single_p_multipatch = splinepy.Multipatch()
    single_p_multipatch.patches = [self_referencing_spline]

    # Determine connectivities
    single_p_multipatch.determine_interfaces()
    multipatch.determine_interfaces()
    assert np.all(
        single_p_multipatch.interfaces == np.array([[1, 0, -1, -1]], dtype=int)
    )

    assert np.all(
        multipatch.interfaces
        == np.array(
            # Global Face IDs and their connectivity
            # [0  1   2   3]  [4   5   6   7]   [8   9  10  11]
            [[-1, 4, 11, -1], [1, -1, -1, -1], [-1, -1, -1, 2]],
            dtype=int,
        )
    )


def test_boundaries(rotated_l_splines):
    """ """
    # init multipatch with multiple splines
    multipatch = splinepy.Multipatch()
    multipatch.patches = rotated_l_splines
    multipatch.determine_interfaces()

    # Using a function
    def west_side(points):
        return points[:, 0] < 0.1

    multipatch.boundary_from_function(west_side)
    assert np.all(
        multipatch.interfaces
        == np.array(
            # Global Face IDs and their connectivity
            # [0  1   2   3]  [4   5   6   7]   [8   9  10  11]
            [[-2, 4, 11, -1], [1, -1, -1, -1], [-2, -1, -1, 2]],
            dtype=int,
        )
    )

    multipatch.set_boundary([0, 1], [3, 3], 3)
    multipatch.set_boundary([1], [1])
    assert np.all(
        multipatch.interfaces
        == np.array(
            # Global Face IDs and their connectivity
            # [0  1   2   3]  [4   5   6   7]   [8   9  10  11]
            [[-2, 4, 11, -3], [1, -4, -1, -3], [-2, -1, -1, 2]],
            dtype=int,
        )
    )

    # Delete all boundaries and determine new ones based on continuity
    multipatch.boundaries_from_continuity()
    assert np.all(
        multipatch.interfaces
        == np.array(
            # Global Face IDs and their connectivity
            # [0  1   2   3]  [4   5   6   7]   [8   9  10  11]
            [[-1, 4, 11, -2], [1, -3, -4, -5], [-1, -6, -7, 2]],
            dtype=int,
        )
    )


def test_interfaces_and_boundaries(are_splines_equal):
    # 2 --- 3 1 --- 0
    # |  1  | |  3  |
    # 0 --- 1 3 --- 2
    # 3 --- 2 0 --- 2
    # |  2  | |  4  |
    # 1 --- 0 1 --- 3
    #
    # with 1 and three flipped in the 3rd axis

    # Create some splines in a random parametric swaps
    b1 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [-2, 0, 3],
            [0, 0, 3],
            [-2, 1, 3],
            [0, 1, 3],
            [-2, 0, 0],
            [0, 0, 0],
            [-2, 1, 0],
            [0, 1, 0],
        ],
    )
    b2 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, -1, 0],
            [-2, -1, 0],
            [0, 0, 0],
            [-2, 0, 0],
            [0, -1, 3],
            [-2, -1, 3],
            [0, 0, 3],
            [-2, 0, 3],
        ],
    )
    b3 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [1, 1, 3],
            [0, 1, 3],
            [1, 0, 3],
            [0, 0, 3],
            [1, 1, 0],
            [0, 1, 0],
            [1, 0, 0],
            [0, 0, 0],
        ],
    )
    b4 = splinepy.Bezier(
        degrees=[1, 1, 1],
        control_points=[
            [0, 0, 0],
            [0, -1, 0],
            [1, 0, 0],
            [1, -1, 0],
            [0, 0, 3],
            [0, -1, 3],
            [1, 0, 3],
            [1, -1, 3],
        ],
    )

    # Multipatch
    multipatch = splinepy.Multipatch([b1, b2, b3, b4])
    multipatch.boundaries_from_continuity()
    assert np.all(
        multipatch.interfaces
        == np.array(
            [
                [-1, 13, 9, -2, -3, -4],
                [20, -1, -5, 2, -4, -3],
                [-6, 1, -2, 18, -3, -4],
                [15, -5, 6, -6, -4, -3],
            ],
            dtype=int,
        )
    )

    # Test boundary multipatch extraction
    default_bmp = multipatch.boundary_multipatch()
    assert len(default_bmp.patches) == 16
    bmp_1 = multipatch.boundary_multipatch(1)
    assert len(bmp_1.patches) == 2
    boundary_1 = []
    for i_patch, i_face in zip(*multipatch.boundaries[0]):
        boundary_1.append(
            *multipatch.patches[i_patch].extract.boundaries([i_face])
        )
    for patch_0, patch_1 in zip(boundary_1, bmp_1.patches):
        assert are_splines_equal(patch_0, patch_1)

    assert len(multipatch.boundary_patch_ids(8)) == 0


def test_add_fields():
    """Test the evaluation and addition to field with setup checks"""
    # Create splines to form multipatch object and "solution" field
    b1 = splinepy.helpme.create.box(1, 2).bspline
    b2 = splinepy.helpme.create.box(2, 2).bspline
    b2.insert_knots(0, [0.2, 0.5])
    b2.cps[:] += [1, 0]

    # 2D field
    two_dim_field = [b1, b2]
    geometry = splinepy.Multipatch([b.create.embedded(3) for b in [b1, b2]])
    # Field dimension is 2 so should raise
    with pytest.raises(RuntimeError) as _:
        geometry.add_fields([two_dim_field], field_dim=1)
    # Success
    geometry.add_fields([two_dim_field], field_dim=2)
    geometry.add_fields([geometry.patches], field_dim=3)
